from __future__ import division, print_function, absolute_import, unicode_literals
#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************
""" Functions for selecting a complete set of fiducials for a GST analysis."""

import numpy as _np
import scipy
from ..tools import frobeniusdist2
from .. import objects as _objs
from .. import construction as _constr
from . import grasp as _grasp
from . import scoring as _scoring


def generate_fiducials(gs_target, omitIdentity=True, eqThresh=1e-6,
                       gatesToOmit=None, forceEmpty=True, maxFidLength=2,
                       algorithm='grasp', algorithm_kwargs=None, verbosity=1):
    """Generate prep and measurement fiducials for a given target gateset.

    Parameters
    ----------
    gs_target : GateSet
        The gateset you are aiming to implement.

    omitIdentity : bool, optional
        Whether to remove the identity gate from the set of gates with which
        fiducials are constructed. Identity gates do nothing to alter
        fiducials, and so should almost always be left out.

    eqThresh : float, optional
        Threshold for determining if a gate is the identity gate. If the square
        Frobenius distance between a given gate and the identity gate is less
        than this threshold, the gate is considered to be an identity gate and
        will be removed from the list of gates from which to construct
        fiducials if `omitIdentity` is ``True``.

    gatesToOmit : list of string, optional
        List of strings identifying gates in the gateset that should not be
        used in fiducials. Oftentimes this will include the identity gate, and
        may also include entangling gates if their fidelity is anticipated to
        be much worse than that of single-system gates.

    forceEmpty : bool, optional (default is True)
        Whether or not to force all fiducial sets to contain the empty gate
        string as a fiducial.

    maxFidLength : int, optional
        The maximum number of gates to include in a fiducial. The default is
        not guaranteed to work for arbitrary gatesets (particularly for quantum
        systems larger than a single qubit).

    algorithm : {'slack', 'grasp'}, optional
        Specifies the algorithm to use to generate the fiducials. Current
        options are:

        'slack'
            See :func:`optimize_integer_fiducials_slack` for more details.
        'grasp'
            Use GRASP to generate random greedy fiducial sets and then locally
            optimize them. See :func:`grasp_fiducial_optimization` for more
            details.

    algorithm_kwargs : dict
        Dictionary of ``{'keyword': keyword_arg}`` pairs providing keyword
        arguments for the specified `algorithm` function. See the documentation
        for functions referred to in the `algorithm` keyword documentation for
        what options are available for each algorithm.

    Returns
    -------
    prepFidList : list of GateString
        A list containing the gate sequences for the prep fiducials.

    measFidList : list of GateString
        A list containing the gate sequences for the measurement fiducials.

    """
    printer = _objs.VerbosityPrinter.build_printer(verbosity)
    if gatesToOmit is None:
        gatesToOmit = []

    fidGates = [gate for gate in gs_target.gates if gate not in gatesToOmit]

    if omitIdentity:
        stateSpaceDims = _np.repeat(1, gs_target.get_dimension())
        stateSpaceLabels = [('L{}'.format(n),)
                            for n in range(gs_target.get_dimension())]
        Identity = _constr.build_gate(stateSpaceDims, stateSpaceLabels,
                                      'I(L0)',
                                      basis=gs_target.get_basis_name())
        for gate in fidGates:
            if frobeniusdist2(gs_target.gates[gate], Identity) < eqThresh:
                fidGates.remove(gate)

    availableFidList = _constr.list_all_gatestrings(fidGates, 0, maxFidLength)

    if algorithm_kwargs is None:
        # Avoid danger of using empty dict for default value.
        algorithm_kwargs = {}

    if algorithm == 'slack':
        printer.log('Using slack algorithm.', 1)
        default_kwargs = {
            'fidList': availableFidList,
            'verbosity': max(0, verbosity - 1),
            'forceEmpty': forceEmpty,
            'scoreFunc': 'all',
            }

        if ('slackFrac' not in algorithm_kwargs
                and 'fixedSlack' not in algorithm_kwargs):
            algorithm_kwargs['slackFrac'] = 1.0
        for key in default_kwargs:
            if key not in algorithm_kwargs:
                algorithm_kwargs[key] = default_kwargs[key]

        prepFidList = optimize_integer_fiducials_slack(gateset=gs_target,
                                                       prepOrMeas='prep',
                                                       **algorithm_kwargs)
        if prepFidList is not None:
            prepScore = compute_composite_score(
                gs_target, prepFidList, 'prep',
                scoreFunc=algorithm_kwargs['scoreFunc'])
            printer.log('Preparation fiducials:', 1)
            printer.log(str([str(fid) for fid in prepFidList]), 1)
            printer.log('Score: {}'.format(prepScore.score), 1)

            
        measFidList = optimize_integer_fiducials_slack(gateset=gs_target,
                                                       prepOrMeas='meas',
                                                       **algorithm_kwargs)
        if measFidList is not None:
            measScore = compute_composite_score(
                gs_target, measFidList, 'meas',
                scoreFunc=algorithm_kwargs['scoreFunc'])
            printer.log('Measurement fiducials:', 1)
            printer.log(str([str(fid) for fid in measFidList]), 1)
            printer.log('Score: {}'.format(measScore.score), 1)

    elif algorithm == 'grasp':
        printer.log('Using GRASP algorithm.', 1)
        default_kwargs = {
            'fidsList': availableFidList,
            'alpha': 0.1,   # No real reason for setting this value of alpha.
            'gatePenalty': 0.1,
            'verbosity': max(0, verbosity - 1),
            'forceEmpty': forceEmpty,
            'scoreFunc': 'all',
            'returnAll': False,
        }
        for key in default_kwargs:
            if key not in algorithm_kwargs:
                algorithm_kwargs[key] = default_kwargs[key]

        prepFidList = grasp_fiducial_optimization(gateset=gs_target,
                                                  prepOrMeas='prep',
                                                  **algorithm_kwargs)

        if algorithm_kwargs['returnAll'] and prepFidList[0] is not None:
            prepScore = compute_composite_score(
                gs_target, prepFidList[0], 'prep',
                scoreFunc=algorithm_kwargs['scoreFunc'])
            printer.log('Preparation fiducials:', 1)
            printer.log(str([str(fid) for fid in prepFidList[0]]), 1)
            printer.log('Score: {}'.format(prepScore.score), 1)
        elif not algorithm_kwargs['returnAll'] and prepFidList is not None:
            prepScore = compute_composite_score(
                gs_target, prepFidList, 'prep',
                scoreFunc=algorithm_kwargs['scoreFunc'])
            printer.log('Preparation fiducials:', 1)
            printer.log(str([str(fid) for fid in prepFidList]), 1)
            printer.log('Score: {}'.format(prepScore.score), 1)

        measFidList = grasp_fiducial_optimization(gateset=gs_target,
                                                  prepOrMeas='meas',
                                                  **algorithm_kwargs)

        if algorithm_kwargs['returnAll'] and measFidList[0] is not None:
            measScore = compute_composite_score(
                gs_target, measFidList[0], 'meas',
                scoreFunc=algorithm_kwargs['scoreFunc'])
            printer.log('Measurement fiducials:', 1)
            printer.log(str([str(fid) for fid in measFidList[0]]), 1)
            printer.log('Score: {}'.format(measScore.score), 1)
        elif not algorithm_kwargs['returnAll'] and measFidList is not None:
            measScore = compute_composite_score(
                gs_target, measFidList, 'meas',
                scoreFunc=algorithm_kwargs['scoreFunc'])
            printer.log('Measurement fiducials:', 1)
            printer.log(str([str(fid) for fid in measFidList]), 1)
            printer.log('Score: {}'.format(measScore.score), 1)

    else:
        raise ValueError("'{}' is not a valid algorithm "
                         "identifier.".format(algorithm))

    return prepFidList, measFidList


#def bool_list_to_ind_list(boolList):
#    output = _np.array([])
#    for i, boolVal in boolList:
#        if boolVal == 1:
#            output = _np.append(i)
#    return output

def xor(*args):
    """Implements logical xor function for arbitrary number of inputs.

    Parameters
    ----------
    args : bool-likes
        All the boolean (or boolean-like) objects to be checked for xor
        satisfaction.

    Returns
    ---------
    output : bool
        True if and only if one and only one element of args is True and the
        rest are False.  False otherwise.

    """

    output = sum(bool(x) for x in args) == 1
    return output

def make_prep_mxs(gs, prepFidList):
    """Make a list of matrices for the gate set preparation operations.

    Makes a list of matrices, where each matrix corresponds to a single
    preparation operation in the gate set, and the column of each matrix is a
    fiducial acting on that state preparation.

    Parameters
    ----------
    gs : GateSet
        The gate set (associates gate matrices with gate labels).

    prepFidList : list of GateStrings
        List of fiducial gate sequences for preparation.

    Returns
    ----------
    outputMatList : list of arrays
        List of arrays, where each array corresponds to one preparation in the
        gate set, and each column therein corresponds to a single fiducial.

    """

    dimRho = gs.get_dimension()
    #numRho = len(gs.preps)
    numFid = len(prepFidList)
    outputMatList = []
    for rho in list(gs.preps.values()):
        outputMat = _np.zeros([dimRho, numFid], float)
        counter = 0
        for prepFid in prepFidList:
            outputMat[:, counter] = _np.dot(gs.product(prepFid), rho).T[0]
            counter += 1
        outputMatList.append(outputMat)
    return outputMatList

def make_meas_mxs(gs, prepMeasList):
    """Make a list of matrices for the gate set measurement operations.

    Makes a list of matrices, where each matrix corresponds to a single
    measurement effect in the gate set, and the column of each matrix is the
    transpose of the measurement effect acting on a fiducial.

    Parameters
    ----------
    gs : GateSet
        The gate set (associates gate matrices with gate labels).

    measFidList : list of GateStrings
        List of fiducial gate sequences for measurement.

    Returns
    ----------
    outputMatList : list of arrays
        List of arrays, where each array corresponds to one measurement in the
        gate set, and each column therein corresponds to a single fiducial.

    """

    dimE = gs.get_dimension()
    # numE = len(gs.effects)
    numFid = len(prepMeasList)
    outputMatList = []
    for E in list(gs.effects.values()):
        outputMat = _np.zeros([dimE, numFid], float)
        counter = 0
        for measFid in prepMeasList:
            outputMat[:, counter] = _np.dot(E.T, gs.product(measFid))[0]
            counter += 1
        outputMatList.append(outputMat)
    return outputMatList


def compute_composite_score(gateset, fidList, prepOrMeas, scoreFunc='all',
                            threshold=1e6, returnAll=False, gatePenalty=0.0,
                            l1Penalty=0.0):
    """Compute a composite score for a fiducial list.

    Parameters
    ----------
    gateset : GateSet
        The gate set (associates gate matrices with gate labels).

    fidList : list of GateStrings
        List of fiducial gate sequences to test.

    prepOrMeas : string ("prep" or "meas")
        Are we testing preparation or measurement fiducials?

    scoreFunc : str ('all' or 'worst'), optional (default is 'all')
        Sets the objective function for scoring a fiducial set.  If 'all',
        score is (number of fiducials) * sum(1/Eigenvalues of score matrix).
        If 'worst', score is (number of fiducials) * 1/min(Eigenvalues of score
        matrix).  Note:  Choosing 'worst' corresponds to trying to make the
        optimizer make the "worst" direction (the one we are least sensitive to
        in Hilbert-Schmidt space) as minimally bad as possible.  Choosing 'all'
        corresponds to trying to make the optimizer make us as sensitive as
        possible to all directions in Hilbert-Schmidt space.  (Also note-
        because we are using a simple integer program to choose fiducials, it
        is possible to get stuck in a local minimum, and choosing one or the
        other objective function can help avoid such minima in different
        circumstances.)

    threshold : float, optional (default is 1e6)
        Specifies a maximum score for the score matrix, above which the
        fiducial set is rejected as informationally incomplete.

    returnAll : bool, optional (default is False)
        Whether the spectrum should be returned along with the score.

    l1Penalty : float, optional (defailt is 0.0)
        Coefficient of a penalty linear in the number of fiducials that is
        added to ``score.score``.

    gatePenalty : float, optional (defailt is 0.0)
        Coefficient of a penalty linear in the total number of gates in all
        fiducials that is added to ``score.score``.

    Returns
    -------
    score : CompositeScore
        The score of the fiducials.

    spectrum : numpy.array, optional
        The eigenvalues of the square of the absolute value of the score
        matrix.

    """
    dimRho = gateset.get_dimension()
    if prepOrMeas == 'prep':
        fidArrayList = make_prep_mxs(gateset, fidList)
    elif prepOrMeas == 'meas':
        fidArrayList = make_meas_mxs(gateset, fidList)
    else:
        raise ValueError('Invalid value "{}" for prepOrMeas (must be "prep" '
                         'or "meas")!'.format(prepOrMeas))
    numMxs = len(fidArrayList)

    numFids = len(fidList)
    scoreMx = _np.zeros([dimRho, numFids *  numMxs], float)
    colInd = 0
    for fidArray in fidArrayList:
        scoreMx[:, colInd:colInd+numFids] = fidArray
        colInd += numFids
    scoreSqMx = _np.dot(scoreMx, scoreMx.T)
    spectrum = sorted(_np.abs(_np.linalg.eigvalsh(scoreSqMx)))
    specLen = len(spectrum)
    N_nonzero = 0
    nonzero_score = _np.inf
    for N in range(1, specLen + 1):
        score = numFids * _scoring.list_score(spectrum[-N:], scoreFunc)
        if score <= 0 or _np.isinf(score) or score > threshold:
            break   # We've found a zero eigenvalue.
        else:
            nonzero_score = score
            N_nonzero = N

    nonzero_score += l1Penalty * len(fidList)

    nonzero_score += gatePenalty * sum([len(fiducial) for fiducial in fidList])

    score = _scoring.CompositeScore(nonzero_score, N_nonzero)

    return (score, spectrum) if returnAll else score


def test_fiducial_list(gateset, fidList, prepOrMeas, scoreFunc='all',
                       returnAll=False, threshold=1e6, l1Penalty=0.0,
                       gatePenalty=0.0):
    """Tests a prep or measure fiducial list for informational completeness.

    Parameters
    ----------
    gateset : GateSet
        The gate set (associates gate matrices with gate labels).

    fidList : list of GateStrings
        List of fiducial gate sequences to test.

    prepOrMeas : string ("prep" or "meas")
        Are we testing preparation or measurement fiducials?

    scoreFunc : str ('all' or 'worst'), optional (default is 'all')
        Sets the objective function for scoring a fiducial set.  If 'all',
        score is (number of fiducials) * sum(1/Eigenvalues of score matrix).
        If 'worst', score is (number of fiducials) * 1/min(Eigenvalues of score
        matrix).  Note:  Choosing 'worst' corresponds to trying to make the
        optimizer make the "worst" direction (the one we are least sensitive to
        in Hilbert-Schmidt space) as minimally bad as possible.  Choosing 'all'
        corresponds to trying to make the optimizer make us as sensitive as
        possible to all directions in Hilbert-Schmidt space.  (Also note-
        because we are using a simple integer program to choose fiducials, it
        is possible to get stuck in a local minimum, and choosing one or the
        other objective function can help avoid such minima in different
        circumstances.)

    returnAll : bool, optional (default is False)
        If true, function returns reciprocals of eigenvalues of fiducial score
        matrix, and the score of the fiducial set as specified by scoreFunc, in
        addition to a boolean specifying whether or not the fiducial set is
        informationally complete

    threshold : float, optional (default is 1e6)
        Specifies a maximum score for the score matrix, above which the
        fiducial set is rejected as informationally incomplete.

    l1Penalty : float, optional (defailt is 0.0)
        Coefficient of a penalty linear in the number of fiducials that is
        added to ``score.score``.

    gatePenalty : float, optional (defailt is 0.0)
        Coefficient of a penalty linear in the total number of gates in all
        fiducials that is added to ``score.score``.

    Returns
    -------
    testResult : bool
        Whether or not the specified fiducial list is informationally complete
        for the provided gate set, to within the tolerance specified by
        threshold.

    spectrum : array, optional
        The number of fiducials times the reciprocal of the spectrum of the
        score matrix.  Only returned if returnAll == True.

    score : float, optional
        The score for the fiducial set; only returned if returnAll == True.

    """

    score, spectrum = compute_composite_score(gateset, fidList, prepOrMeas,
                                              scoreFunc=scoreFunc,
                                              threshold=threshold,
                                              returnAll=True,
                                              l1Penalty=l1Penalty,
                                              gatePenalty=gatePenalty)

    if score.N < len(spectrum):
        testResult = False
    else:
        testResult = True

    return (testResult, spectrum, score) if returnAll else testResult


def build_bitvec_mx(n, k):
    """Create an array of all fixed length and Hamming weight binary vectors.

    Parameters
    ----------
    n : int
        The length of each bit string.
    k : int
        The hamming weight of each bit string.

    Returns
    ----------
    bitVecMx : _np array
        this is the array of binary vectors of a fixed length n and fixed
        Hamming weight k.

    """
    bitVecMx = _np.zeros([int(scipy.special.binom(n, k)), n])
    diff = n - k

    # Recursive function for populating a matrix of arbitrary size
    def build_mx(previous_bit_locs, i, counter):
        """Allows arbitrary nesting of for loops

        Parameters
        ----------
        previous_bit_locs : tuple
            current loop contents, ex:

            >>> for i in range(10):
            >>>    for j in range(10):
            >>>        (i, j)

        i : int
            Loop depth

        counter : int
            tracks which fields of mx have been already set

        Returns
        ----------
        counter : int
            for updating the counter one loop above the current one

        """
        if i == 0:
            bitVecMx[counter][list(previous_bit_locs)] = 1
            counter += 1
        else:
            subK = k - i
            # Recursive definition allowing arbitrary size
            last_bit_loc = previous_bit_locs[-1] # More explicit?
            for bit_loc in range(1+last_bit_loc, diff+subK+1):
                current_bit_locs = previous_bit_locs + (bit_loc,)

                counter = build_mx(current_bit_locs, i - 1, counter)

        # An alternative to shared state
        return counter

    counter = 0
    for bit_loc_0 in range(diff+1):
        counter = build_mx((bit_loc_0,), k - 1, counter) # Do subK additional iterations

    return bitVecMx

def optimize_integer_fiducials_slack(gateset, fidList, prepOrMeas=None,
                                     initialWeights=None, scoreFunc='all',
                                     maxIter=100, fixedSlack=None,
                                     slackFrac=None, returnAll=False,
                                     forceEmpty=True, forceEmptyScore=1e100,
                                     fixedNum=None, threshold=1e6,
                                     # forceMinScore=1e100,
                                     verbosity=1):
    """Find a locally optimal subset of the fiducials in fidList.

    Locally optimal here means that no single fiducial can be excluded without
    increasing the sum of the reciprocals of the singular values of the "score
    matrix" (the matrix whose columns are the fiducials acting on the
    preparation, or the transpose of the measurement acting on the fiducials),
    by more than a fixed or variable amount of "slack", as specified by
    fixedSlack or slackFrac.

    Parameters
    ----------
    gateset : GateSet
        The gate set (associates gate matrices with gate labels).

    fidList : list of GateStrings
        List of all fiducials gate sequences to consider.

    initialWeights : list-like
        List or array of either booleans or (0 or 1) integers specifying which
        fiducials in fidList comprise the initial fiduial set.  If None, then
        starting point includes all fiducials.

    scoreFunc : str ('all' or 'worst'), optional (default is 'all')
        Sets the objective function for scoring a fiducial set.  If 'all',
        score is (number of fiducials) * sum(1/Eigenvalues of score matrix).
        If 'worst', score is (number of fiducials) * 1/min(Eigenvalues of score
        matrix).  Note:  Choosing 'worst' corresponds to trying to make the
        optimizer make the "worst" direction (the one we are least sensitive to
        in Hilbert-Schmidt space) as minimally bad as possible.  Choosing 'all'
        corresponds to trying to make the optimizer make us as sensitive as
        possible to all directions in Hilbert-Schmidt space.  (Also note-
        because we are using a simple integer program to choose fiducials, it
        is possible to get stuck in a local minimum, and choosing one or the
        other objective function can help avoid such minima in different
        circumstances.)

    maxIter : int, optional
        The maximum number of iterations before stopping.

    fixedSlack : float, optional
        If not None, a floating point number which specifies that excluding a
        fiducial is allowed to increase the fiducial set score additively by
        fixedSlack.  You must specify *either* fixedSlack or slackFrac.

    slackFrac : float, optional
        If not None, a floating point number which specifies that excluding a
        fiducial is allowed to increase the fiducial set score multiplicatively
        by (1+slackFrac).  You must specify *either* fixedSlack or slackFrac.

    returnAll : bool, optional
        If True, return the final "weights" vector and score dictionary in
        addition to the optimal fiducial list (see below).

    forceEmpty : bool, optional (default is True)
        Whether or not to force all fiducial sets to contain the empty gate
        string as a fiducial.

        IMPORTANT:  This only works if the first element of fidList is the
        empty gate string.

    forceEmptyScore : float, optional (default is 1e100)
        When forceEmpty is True, what score to assign any fiducial set that
        does not contain the empty gate string as a fiducial.

    forceMin : bool, optional (default is False)
        If True, forces fiducial selection to choose a fiducial set that is *at
        least* as large as forceMinNum.

    forceMinNum : int, optional (default is None)
        If not None, and forceMin == True, the minimum size of the returned
        fiducial set.

    forceMinScore : float, optional (default is 1e100)
        When forceMin is True, what score to assign any fiducial set that does
        not contain at least forceMinNum fiducials.

    threshold : float, optional (default is 1e6)
        Entire fiducial list is first scored before attempting to select
        fiducials; if score is above threshold, then fiducial selection will
        auto-fail.  If final fiducial set selected is above threshold, then
        fiducial selection will print a warning, but return selected set.

    verbosity : int, optional
        Integer >= 0 indicating the amount of detail to print.

    Returns
    -------
    finalFidList : list
        Sublist of fidList specifying the final, optimal, set of fiducials.

    weights : array
        Integer array, of length len(fidList), containing 0s and 1s to indicate
        which elements of fidList were chosen as finalFidList.  Only returned
        when returnAll == True.

    scoreDictionary : dict
        Dictionary with keys == tuples of 0s and 1s of length len(fidList),
        specifying a subset of fiducials, and values == 1.0/smallest-non-gauge-
        eigenvalue "scores".

    """
    printer = _objs.VerbosityPrinter.build_printer(verbosity)

    if not xor(fixedSlack, slackFrac):
        raise ValueError("One and only one of fixedSlack or slackFrac should "
                         "be specified!")

    initial_test = test_fiducial_list(gateset, fidList, prepOrMeas,
                                      scoreFunc=scoreFunc, returnAll=True,
                                      threshold=threshold)
    if initial_test[0]:
        printer.log("Complete initial fiducial set succeeds.", 1)
        printer.log("Now searching for best fiducial set.", 1)
    else:
        printer.warning("Complete initial fiducial set FAILS.")
        printer.warning("Aborting search.")
        return None

    #Initially allow adding to weight. -- maybe make this an argument??
    lessWeightOnly = False

    nFids = len(fidList)

    dimRho = gateset.get_dimension()

    printer.log("Starting fiducial set optimization. Lower score is better.",
                1)

    scoreD = {}

    #fidLengths = _np.array( list(map(len,fidList)), 'i')
    if prepOrMeas == 'prep':
        fidArrayList = make_prep_mxs(gateset, fidList)
    elif prepOrMeas == 'meas':
        fidArrayList = make_meas_mxs(gateset, fidList)
    else:
        raise Exception('prepOrMeas must be specified!')
    numMxs = len(fidArrayList)

    def compute_score(wts, cache_score=True):
        score = None
        if forceEmpty and _np.count_nonzero(wts[:1]) != 1:
            score = forceEmptyScore
#        if forceMinNum and _np.count_nonzero(wts) < forceMinNum:
#            score = forceMinScore
        if score is None:
            numFids = _np.sum(wts)
            scoreMx = _np.zeros([dimRho, int(numFids) *  int(numMxs)], float)
            colInd = 0
            wts = _np.array(wts)
            wtsLoc = _np.where(wts)[0]
            for fidArray in fidArrayList:
                scoreMx[:, colInd:colInd+int(numFids)] = fidArray[:, wtsLoc]
                colInd += int(numFids)
            scoreSqMx = _np.dot(scoreMx, scoreMx.T)
#            score = numFids * _np.sum(1./_np.linalg.eigvalsh(scoreSqMx))
            score = numFids * _scoring.list_score(
                _np.linalg.eigvalsh(scoreSqMx), scoreFunc)
            if score <= 0 or _np.isinf(score):
                score = 1e10
        if cache_score:
            scoreD[tuple(wts)] = score
        return score

    if fixedNum is not None:
        if forceEmpty:
            hammingWeight = fixedNum - 1
            numBits = len(fidList) - 1
        else:
            hammingWeight = fixedNum
            numBits = len(fidList)
        numFidLists = scipy.special.binom(numBits, hammingWeight)
        printer.log("Output set is required to be of size%s" % fixedNum, 1)
        printer.log("Total number of fiducial sets to be checked is%s"
                    % numFidLists, 1)
        printer.warning("If this is very large, you may wish to abort.")
#        print "Num bits:", numBits
#        print "Num Fid Options:", hammingWeight
        # Now a non auxillary function:
        bitVecMat = build_bitvec_mx(numBits, hammingWeight)

        if forceEmpty:
            bitVecMat = _np.concatenate((_np.array([[1]*int(numFidLists)]).T,
                                         bitVecMat), axis=1)
        best_score = _np.inf
        # Explicitly declare best_weights, even if it will soon be replaced
        best_weights = []
        for weights in bitVecMat:
            temp_score = compute_score(weights, cache_score=True)
            # If scores are within machine precision, we want the fiducial set
            # that requires fewer total button gate operations.
            if abs(temp_score - best_score) < 1e-8:
#                print "Within machine precision!"
                bestFidList = []
                for index, val in enumerate(best_weights):
                    if val == 1:
                        bestFidList.append(fidList[index])
                tempFidList = []
                for index, val in enumerate(weights):
                    if val == 1:
                        tempFidList.append(fidList[index])
                tempLen = sum(len(i) for i in tempFidList)
                bestLen = sum(len(i) for i in bestFidList)
#                print tempLen, bestLen
#                print temp_score, best_score
                if tempLen < bestLen:
                    best_score = temp_score
                    best_weights = weights
                    printer.log("Switching!", 1)
            elif temp_score < best_score:
                best_score = temp_score
                best_weights = weights

        goodFidList = []
        weights = best_weights
        for index, val in enumerate(weights):
            if val == 1:
                goodFidList.append(fidList[index])

        if returnAll:
            return goodFidList, weights, scoreD
        else:
            return goodFidList


        return best_score, best_weights, finalFidList #pylint: disable=undefined-variable

    def get_neighbors(boolVec):
        for i in range(nFids):
            v = boolVec.copy()
            v[i] = (v[i] + 1) % 2 #toggle v[i] btwn 0 and 1
            yield v

    if initialWeights is not None:
        weights = _np.array([1 if x else 0 for x in initialWeights])
    else:
        weights = _np.ones(nFids, 'i') #default: start with all germs
        lessWeightOnly = True #we're starting at the max-weight vector

    score = compute_score(weights)
    L1 = sum(weights) # ~ L1 norm of weights

    with printer.progress_logging(1):

        for iIter in range(maxIter):
            scoreD_keys = scoreD.keys() #list of weight tuples already computed

            printer.show_progress(iIter, maxIter,
                                  suffix="score=%g, nFids=%d" % (score, L1))

            bFoundBetterNeighbor = False
            for neighbor in get_neighbors(weights):
                if tuple(neighbor) not in scoreD_keys:
                    neighborL1 = sum(neighbor)
                    neighborScore = compute_score(neighbor)
                else:
                    neighborL1 = sum(neighbor)
                    neighborScore = scoreD[tuple(neighbor)]

                # Move if we've found better position; if we've relaxed, we
                # only move when L1 is improved.
                if neighborScore <= score and (neighborL1 < L1
                                               or not lessWeightOnly):
                    weights, score, L1 = neighbor, neighborScore, neighborL1
                    bFoundBetterNeighbor = True
                    printer.log("Found better neighbor: nFids = %d score = %g"
                                % (L1, score), 3)


            if not bFoundBetterNeighbor: # Time to relax our search.
                # from now on, don't allow increasing weight L1
                lessWeightOnly = True

                if fixedSlack:
                    # Note score is positive (for sum of 1/lambda)
                    slack = score + fixedSlack
                elif slackFrac:
                    slack = score * slackFrac
                assert slack > 0

                printer.log("No better neighbor. "
                            "Relaxing score w/slack: %g => %g"
                            % (score, score+slack), 2)
                # artificially increase score and see if any neighbor is better
                # now...
                score += slack

                for neighbor in get_neighbors(weights):
                    if sum(neighbor) < L1 and scoreD[tuple(neighbor)] < score:
                        weights, score, L1 = (neighbor,
                                              scoreD[tuple(neighbor)],
                                              sum(neighbor))
                        bFoundBetterNeighbor = True
                        printer.log("Found better neighbor: nFids = %d "
                                    "score = %g" % (L1, score), 3)

                if not bFoundBetterNeighbor: #Relaxing didn't help!
                    printer.log("Stationary point found!", 2)
                    break # end main for loop

            printer.log("Moving to better neighbor", 2)
        else:
            printer.log("Hit max. iterations", 2)

    printer.log("score = %s" % score, 1)
    printer.log("weights = %s" % weights, 1)
    printer.log("L1(weights) = %s" % sum(weights), 1)

    goodFidList = []
    for index, val in enumerate(weights):
        if val == 1:
            goodFidList.append(fidList[index])

    # final_test = test_fiducial_list(gateset, goodFidList, prepOrMeas,
    #                                 scoreFunc=scoreFunc, returnAll=True,
    #                                 threshold=threshold)
    if initial_test[0]:
        printer.log("Final fiducial set succeeds.", 1)
    else:
        printer.log("WARNING: Final fiducial set FAILS.", 1)

    if returnAll:
        return goodFidList, weights, scoreD
    else:
        return goodFidList


def grasp_fiducial_optimization(gateset, fidsList, prepOrMeas, alpha,
                                iterations=5, scoreFunc='all', gatePenalty=0.0,
                                l1Penalty=0.0, returnAll=False,
                                forceEmpty=True, threshold=1e6, seed=None,
                                verbosity=0):
    """Use GRASP to find a high-performing set of fiducials.

    """
    printer = _objs.VerbosityPrinter.build_printer(verbosity)

    if prepOrMeas not in ['prep', 'meas']:
        raise ValueError("'{}' is an invalid value for prepOrMeas (must be "
                         "'prep' or 'meas')!".format(prepOrMeas))

    initial_test = test_fiducial_list(gateset, fidsList, prepOrMeas,
                                      scoreFunc=scoreFunc, returnAll=False,
                                      threshold=threshold)
    if initial_test:
        printer.log("Complete initial fiducial set succeeds.", 1)
        printer.log("Now searching for best fiducial set.", 1)
    else:
        printer.warning("Complete initial fiducial set FAILS.")
        printer.warning("Aborting search.")
        return (None, None, None) if returnAll else None

    initialWeights = _np.zeros(len(fidsList), dtype='i')
    if forceEmpty:
        fidsLens = [len(fiducial) for fiducial in fidsList]
        initialWeights[fidsLens.index(0)] = 1

    getNeighborsFn = lambda weights: _grasp.get_swap_neighbors(
        weights, forcedWeights=initialWeights)

    printer.log("Starting fiducial list optimization. Lower score is better.",
                1)

    # Dict of keyword arguments passed to compute_score_non_AC that don't
    # change from call to call
    compute_kwargs = {
        'gateset': gateset,
        'prepOrMeas': prepOrMeas,
        'scoreFunc': scoreFunc,
        'threshold': threshold,
        'gatePenalty': gatePenalty,
        'returnAll': False,
        'l1Penalty': 0.0,
        }

    final_compute_kwargs = compute_kwargs.copy()
    final_compute_kwargs['l1Penalty'] = l1Penalty

    scoreFn = lambda fidList: compute_composite_score(fidList=fidList,
                                                      **compute_kwargs)

    finalScoreFn = lambda fidList: compute_composite_score(
        fidList=fidList, **final_compute_kwargs)

    dimRho = gateset.get_dimension()
    feasibleThreshold=_scoring.CompositeScore(threshold, dimRho)

    rclFn = lambda x: _scoring.composite_rcl_fn(x, alpha)

    initialSolns = []
    localSolns = []

    for iteration in range(iterations):
        # This loop is parallelizable (each iteration is independent of all
        # other iterations).
        printer.log('Starting iteration {} of {}.'.format(iteration + 1,
                                                          iterations), 1)
        success = False
        failCount = 0
        while not success and failCount < 10:
            try:
                iterSolns = _grasp.do_grasp_iteration(
                    elements=fidsList, greedyScoreFn=scoreFn, rclFn=rclFn,
                    localScoreFn=scoreFn,
                    getNeighborsFn=getNeighborsFn,
                    feasibleThreshold=feasibleThreshold,
                    initialElements=initialWeights, seed=seed,
                    verbosity=verbosity)

                initialSolns.append(iterSolns[0])
                localSolns.append(iterSolns[1])

                success = True
                printer.log('Finished iteration {} of {}.'.format(
                    iteration + 1, iterations), 1)
            except Exception as e:
                failCount += 1
                if failCount == 10:
                    raise e
                else:
                    printer.warning(e)

    finalScores = _np.array([finalScoreFn(localSoln)
                             for localSoln in localSolns])
    bestSoln = localSolns[_np.argmin(finalScores)]

    return (bestSoln, initialSolns, localSolns) if returnAll else bestSoln



from __future__ import division, print_function, absolute_import, unicode_literals
#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************
"""Defines the Profiler class and supporting functionality"""

import time as _time
import os as _os
import inspect as _inspect
import itertools as _itertools
import numpy as _np

try:
    import psutil as _psutil
    def _get_mem_usage():
        p = _psutil.Process(_os.getpid())
        return p.memory_info()[0]

except ImportError:
    import sys as _sys
    import resource as _resource
    def _get_mem_usage():
        mem = _resource.getrusage(_resource.RUSAGE_SELF).ru_maxrss
          # peak memory usage (bytes on OS X, kilobytes on Linux)
        if _sys.platform != 'darwin': mem *= 1024 #now always in bytes
        return mem

def _get_root_mem_usage(comm):
    """ Returns the memory usage on the 0th processor """
    mem = _get_mem_usage()
    if comm is not None: 
        if comm.Get_rank() == 0:
            comm.bcast(mem,root=0)
        else:
            mem = comm.bcast(None, root=0)
    return mem

def _get_max_mem_usage(comm):
    """ Returns the memory usage on the 0th processor """
    mem = _get_mem_usage()
    if comm is not None: 
        memlist = comm.allgather(mem)
        mem = max(memlist)
    return mem


BtoGB = 1.0/(1024.0**3) #convert bytes -> GB

class Profiler(object):
    """
    Profiler objects are used for tracking both time and memory usage.
    """

    def __init__(self,comm=None,default_print_memcheck=False):
        """
        Construct a new Profiler instance.

        Parameters
        ----------
        comm : mpi4py.MPI.Comm, optional
           MPI communicator so only profile and print messages on root proc.
        """
        self.comm = comm
        self.timers = {}
        self.counters = {}
        self.mem_checkpoints = {}
        self.print_memchecks = default_print_memcheck

    def add_time(self, name, start_time, prefix=0):
        """
        Adds an elapsed time to a named "timer"-type accumulator.
        
        Parameters
        ----------
        name : string
           The name of the timer to add elapsed time into (if the name doesn't
           exist, one is created and initialized to the elapsed time).
           
        start_time : float
           The starting time used to compute the elapsed, i.e. the value 
           `time.time()-start_time`, which is added to the named timer.

        prefix : int, optional
           Prefix to the timer name the current stack depth and this number
           of function names, starting with the current function and moving
           the call stack.  When zero, no prefix is added. For example, 
           with `prefix == 1`, "Total" might map to " 3: myFunc: Total".

        Returns
        -------
        None
        """
        if prefix > 0:
            stack = _inspect.stack()
            try:
                depth = len(stack)-1 # -1 to discount current fn (add_time)
                functions = " : ".join(_inspect.getframeinfo(frm[0]).filename
                                       for frm in reversed(stack[1:1+prefix]))
                name = "%2d: %s: %s" % (depth,functions,name)
            finally:
                stack = None #make sure frames get cleaned up properly

        val = _time.time() - start_time
        if name in self.timers:
            self.timers[name] += val
        else:
            self.timers[name] = val


    def add_count(self, name, inc=1, prefix=0):
        """
        Adds a given value to a named "counter"-type accumulator.
        
        Parameters
        ----------
        name : string
           The name of the counter to add `val` into (if the name doesn't exist,
           one is created and initialized to `val`).
           
        inc : int, optional
           The increment (the value to add to the counter).

        prefix : int, optional
           Prefix to the timer name the current stack depth and this number
           of function names, starting with the current function and moving
           the call stack.  When zero, no prefix is added. For example, 
           with `prefix == 1`, "Total" might map to " 3: myFunc: Total".

        Returns
        -------
        None
        """
        if prefix > 0:
            stack = _inspect.stack()
            try:
                depth = len(stack)-1 # -1 to discount current fn (add_count)
                functions = " : ".join(_inspect.getframeinfo(frm[0]).filename
                                       for frm in reversed(stack[1:1+prefix]))
                name = "%2d: %s: %s" % (depth,functions,name)
            finally:
                stack = None #make sure frames get cleaned up properly

        if name in self.counters:
            self.counters[name] += inc
        else:
            self.counters[name] = inc


    def mem_check(self, name, printme=None, prefix=0):
        """
        Record the memory usage at this point and tag with a `name`.
        
        Parameters
        ----------
        name : string
           The name of the memory checkpoint.  (Later, memory information can
           be organized by checkpoint name.)
           
        printme : bool, optional
           Whether or not to print the memory usage during this function call
           (if None, the default, then the value of `default_print_memcheck`
           specified during Profiler construction is used).

        prefix : int, optional
           Prefix to the timer name the current stack depth and this number
           of function names, starting with the current function and moving
           the call stack.  When zero, no prefix is added. For example, 
           with `prefix == 1`, "Total" might map to " 3: myFunc: Total".

        Returns
        -------
        None
        """
        if prefix > 0:
            stack = _inspect.stack()
            try:
                depth = len(stack)-1 # -1 to discount current fn (mem_check)
                functions = " : ".join(_inspect.getframeinfo(frm[0]).filename
                                       for frm in reversed(stack[1:1+prefix]))
                name = "%2d: %s: %s" % (depth,functions,name)
            finally:
                stack = None #make sure frames get cleaned up properly

        usage = _get_mem_usage()
        timestamp = _time.time()
        if name in self.mem_checkpoints:
            self.mem_checkpoints[name].append( (timestamp,usage) )
        else:
            self.mem_checkpoints[name] = [ (timestamp,usage) ]

        bPrint = self.print_memchecks if (printme is None) else printme
        if bPrint: self.print_mem(name)

    def print_mem(self, name, show_minmax=False):
        """
        Prints the current memory usage (but doesn't store it).

        Useful for debugging, this function prints the current memory
        usage - optionally giving the mininum, maximum, and average
        across all the processors.

        Parameters
        ----------
        name : string
            A label to print before the memory usage number(s).

        show_minmax : bool, optional
            If True and there are multiple processors, print the
            min, average, and max memory usage from among the processors.
            Note that this will invoke MPI collective communication and so
            this `print_mem` call **must** be executed by all the processors.
            If False and there are multiple processors, only the rank 0
            processor prints output.

        Returns
        -------
        None
        """
        usage = _get_mem_usage()
        if self.comm is not None:
            if show_minmax:
                memlist = self.comm.gather(usage, root=0)
                if self.comm.Get_rank() == 0:
                    avg_usage = sum(memlist)*BtoGB / self.comm.Get_size()
                    min_usage = min(memlist)*BtoGB
                    max_usage = max(memlist)*BtoGB
                    print("MEM USAGE [%s] = %.2f GB, %.2f GB, %.2f GB" %
                          (name,min_usage,avg_usage,max_usage))
            elif self.comm.Get_rank() == 0:
                print("MEM USAGE [%s] = %.2f GB" % (name,usage*BtoGB))
        else:
            print("MEM USAGE [%s] = %.2f GB" % (name,usage*BtoGB))


    def print_msg(self, msg, all_ranks=False):
        """ 
        Prints a message to stdout, possibly from all ranks.

        A utility function used in debugging, this function offers a 
        convenient way to print a message on only the root processor 
        or on all processors.

        Parameters
        ----------
        msg : string
            The message to print.

        all_ranks : bool, optional
            If True, all processors will print `msg`, preceded by their
            rank label (e.g. "Rank4: ").  If False, only the rank 0
            processor will print the message.

        Returns
        -------
        None
        """
        if self.comm is not None:
            if all_ranks:
                print("Rank%d: %s" % (self.comm.Get_rank(),msg))
            elif self.comm.Get_rank() == 0:
                print(msg)
        else: print(msg)


    def format_times(self, sortBy="name"):
        """
        Formats a string to report the timer values recorded in this Profiler.

        Parameters
        ----------
        sortBy : {"name","time"}
            What to sort list of timers by.

        Returns
        -------
        str
        """
        s = "---> Times (by %s): \n" % sortBy
        if sortBy == "name":
            timerNames = sorted(list(self.timers.keys()))
        elif sortBy == "time":
            timerNames = sorted(list(self.timers.keys()),
                                key=lambda x: self.timers[x])
        else:
            raise ValueError("Invalid 'sortBy' argument: %s" % sortBy)
        
        for nm in timerNames:
            s += "  %s : %.1fs\n" % (nm,self.timers[nm])
        s += "\n"
        return s


    def format_counts(self, sortBy="name"):
        """
        Formats a string to report the counter values recorded in this Profiler.

        Parameters
        ----------
        sortBy : {"name","count"}
           What to sort list of counts by.

        Returns
        -------
        str
        """
        s = "---> Counters (by %s): \n" % sortBy
        if sortBy == "name":
            counterNames = sorted(list(self.counters.keys()))
        elif sortBy == "count":
            counterNames = sorted(list(self.counters.keys()),
                                key=lambda x: self.counters[x])
        else:
            raise ValueError("Invalid 'sortBy' argument: %s" % sortBy)
        
        for nm in counterNames:
            s += "  %s : %d\n" % (nm,self.counters[nm])
        s += "\n"
        return s


    def format_memory(self, sortBy="name"):
        """
        Formats a string to report the memory usage checkpoints recorded
        in this Profiler.

        Parameters
        ----------
        sortBy : {"name","usage","timestamp"}
           What to sort list of counts by.

        Returns
        -------
        str
        """
        if len(self.mem_checkpoints) == 0:
            return "No memory checkpoints"

        #for key in self.mem_checkpoints:
        #    print("ITEM:",self.mem_checkpoints[key])
        #    assert(False)
        #print("LIST: ",list(self.mem_checkpoints.values()))
        max_memory = max([ usage for timestamp,usage in 
                           _itertools.chain(*self.mem_checkpoints.values())])
        s  = "---> Max Memory usage = %.2fGB\n" % (max_memory*BtoGB)
        s += "---> Memory usage (by %s): \n" % sortBy

        if sortBy == "timestamp": #special case in that we print each event,
                                  # not just the average usage per checkpoint
            raise NotImplementedError("TODO")

        avg_usages = { k : _np.mean([u for t,u in infos]) for k,infos 
                       in self.mem_checkpoints.items() }

        if sortBy == "name":
            chkptNames = sorted(list(self.mem_checkpoints.keys()))
        elif sortBy == "usage":
            chkptNames = sorted(list(avg_usages.keys()),
                                key=lambda x: avg_usages[x])
        else:
            raise ValueError("Invalid 'sortBy' argument: %s" % sortBy)
        
        for nm in chkptNames:
            usages = [u for t,u in self.mem_checkpoints[nm]]
            s += "  %s : %.2fGB (min=%.2f,max=%.2f)\n" % \
                (nm,avg_usages[nm]*BtoGB,min(usages)*BtoGB,max(usages)*BtoGB)
        s += "\n"
        return s

        

    def __getstate__(self):
        #Return the state (for pickling) -- *don't* pickle Comm object
        to_pickle = self.__dict__.copy()
        del to_pickle['comm'] # one *cannot* pickle Comm objects
        return  to_pickle

    def __setstate__(self, stateDict):
        self.__dict__.update(stateDict)
        self.comm = None # initialize to None upon unpickling

        


class DummyProfiler(object):
    """
    A class which implements the same interface as Profiler but 
    which doesn't actually do any profiling (consists of stub functions).
    """

    def __init__(self):
        """
        Construct a new DummyProfiler instance.
        """

    def add_time(self, name, start_time):
        """Stub function that does nothing"""
        pass

    def add_count(self, name, inc=1):
        """Stub function that does nothing"""
        pass

    def mem_check(self, name,printme=None):
        """Stub function that does nothing"""
        pass

    



from __future__  import division, print_function, absolute_import, unicode_literals
from .formatter  import FormatSet   as _FormatSet
from collections import OrderedDict as _OrderedDict
import re as _re

class ReportTable(object):
    def __init__(self, colHeadings, formatters, customHeader=None):
        self._headings          = colHeadings
        self._headingFormatters = formatters
        self._customHeadings    = customHeader
        self._rows              = []

        if self._headingFormatters is not None:
            self._columnNames = self._headings
        else: #headingFormatters is None => headings is dict w/formats
            self._columnNames = self._headings['text'] #use text heading


    def addrow(self, rowData, formatters):
        self._rows.append((rowData, formatters))

    def finish(self):
        pass #nothing to do currently

    def render(self, fmt, longtables=False, tableclass='pygstiTbl',
               scratchDir=None, precision=6, polarprecision=3, sciprecision=0):

        specs = {
            'scratchDir'     : scratchDir,
            'precision'      : precision,
            'polarprecision' : polarprecision,
            'sciprecision'   : sciprecision
            }

        # Create a formatSet, which contains rules for rendering lists
        formatSet =  _FormatSet(specs)

        if fmt == "latex":

            table = "longtable" if longtables else "tabular"
            if self._customHeadings is not None \
                    and "latex" in self._customHeadings:
                latex = self._customHeadings['latex']
            else:
                if self._headingFormatters is not None:
                    colHeadings_formatted = formatSet.formatList(
                                             self._headings,
                                             self._headingFormatters, "latex")
                else: #headingFormatters is None => headings is dict w/formats
                    colHeadings_formatted = self._headings['latex']

                latex  = "\\begin{%s}[l]{%s}\n\hline\n" % \
                    (table, "|c" * len(colHeadings_formatted) + "|")
                latex += ("%s \\\\ \hline\n"
                          % (" & ".join(colHeadings_formatted)))

            for rowData, formatters in self._rows:
                formatted_rowData = formatSet.formatList(rowData, formatters,
                                                          "latex")
                if len(formatted_rowData) > 0:
                    latex += " & ".join(formatted_rowData)

                    multirows = [ ("multirow" in el) for el in formatted_rowData ]
                    if any(multirows):
                        latex += " \\\\ "
                        last = True; lineStart = None; col = 1
                        for multi,data in zip(multirows,formatted_rowData):                                
                            if last == True and multi == False:
                                lineStart = col #line start
                            elif last == False and multi == True:
                                latex += "\cline{%d-%d} " % (lineStart,col) #line end
                            last=multi
                            res = _re.search("multicolumn{([0-9])}",data)
                            if res: col += int(res.group(1))
                            else:   col += 1
                        if last == False: #need to end last line
                            latex += "\cline{%d-%d} "%(lineStart,col-1)
                        latex += "\n"
                    else:
                       latex += " \\\\ \hline\n"

            latex += "\end{%s}\n" % table
            return latex


        elif fmt == "html":

            if self._customHeadings is not None \
                    and "html" in self._customHeadings:
                html = self._customHeadings['html']
            else:
                if self._headingFormatters is not None:
                    colHeadings_formatted = \
                        formatSet.formatList(self._headings,
                                       self._headingFormatters, "html")
                else: #headingFormatters is None => headings is dict w/formats
                    colHeadings_formatted = self._headings['html']

                html  = "<table class=%s><thead>" % tableclass
                html += "<tr><th> %s </th></tr>" % \
                    (" </th><th> ".join(colHeadings_formatted))
                html += "</thead><tbody>"

            for rowData,formatters in self._rows:
                formatted_rowData = formatSet.formatList(rowData, formatters, "html")
                if len(formatted_rowData) > 0:
                    html += "<tr><td>" + \
                        "</td><td>".join(formatted_rowData) + "</td></tr>\n"

            html += "</tbody></table>"
            return html


        elif fmt == 'text':

            if self._customHeadings is not None \
                    and 'text' in self._customHeadings:
                raise ValueError("custom headers unsupported for text format")

            if self._headingFormatters is not None:
                colHeadings_formatted = \
                    formatSet.formatList(self._headings,
                                   self._headingFormatters, 'text')
            else: #headingFormatters is None => headings is dict w/formats
                colHeadings_formatted = self._headings['text']

            text = { 'column names': colHeadings_formatted,
                   'row data': [] }

            for rowData,formatters in self._rows:
                print(rowData)
                formatted_rowData = formatSet.formatList(rowData, formatters, 'text')
                if len(formatted_rowData) > 0:
                    text['row data'].append( formatted_rowData )

            return text


        elif fmt == "ppt":

            if self._customHeadings is not None \
                    and "ppt" in self._customHeadings:
                raise ValueError("custom headers unsupported for " +
                                 "powerpoint format")

            if self._headingFormatters is not None:
                colHeadings_formatted = \
                    formatSet.formatList(self._headings,
                                   self._headingFormatters, "ppt")
            else: #headingFormatters is None => headings is dict w/formats
                colHeadings_formatted = self._headings['ppt']

            ppt = { 'column names': colHeadings_formatted,
                    'row data' : [] }

            for rowData,formatters in self._rows:
                formatted_rowData = formatSet.formatList(rowData, formatters, "ppt")
                if len(formatted_rowData) > 0:
                    ppt['row data'].append( formatted_rowData )

            return ppt

        else:
            raise ValueError("Unknown format: %s" % fmt)


    def __str__(self):

        def strlen(x):
            return max([len(p) for p in str(x).split('\n')])
        def nlines(x):
            return len(str(x).split('\n'))
        def getline(x,i):
            lines = str(x).split('\n')
            return lines[i] if i < len(lines) else ""

        self.render('text')
        col_widths = [0]*len(self._columnNames)
        row_lines = [0]*len(self._rows)
        header_lines = 0

        for i,nm in enumerate(self._columnNames):
            col_widths[i] = max( strlen(nm), col_widths[i] )
            header_lines = max(header_lines, nlines(nm))
        for k,(d,_) in enumerate(self._rows):
            for i,el in enumerate(d):
                col_widths[i] = max( strlen(el), col_widths[i] )
                row_lines[k] = max(row_lines[k], nlines(el))

        row_separator = "|" + '-'*(sum([w+5 for w in col_widths])-1) + "|\n"
          # +5 for pipe & spaces, -1 b/c don't count first pipe

        s  = "*** ReportTable object ***\n"
        s += row_separator

        for k in range(header_lines):
            for i,nm in enumerate(self._columnNames):
                s += "|  %*s  " % (col_widths[i],getline(nm,k))
            s += "|\n"
        s += row_separator

        for rowIndex,(rowEls,_) in enumerate(self._rows):
            for k in range(row_lines[rowIndex]):
                for i,el in enumerate(rowEls):
                    s += "|  %*s  " % (col_widths[i],getline(el,k))
                s += "|\n"
            s += row_separator

        s += "\n"
        s += "Access row and column data by indexing into this object\n"
        s += " as a dictionary using the column header followed by the\n"
        s += " value of the first element of each row, i.e.,\n"
        s += " tableObj[<column header>][<first row element>].\n"

        return s


    def __getitem__(self, key):
        """Indexes the first column rowdata"""
        for row_data,_ in self._rows:
            if len(row_data) > 0 and row_data[0] == key:
                return _OrderedDict( zip(self._columnNames,row_data) )
        raise KeyError("%s not found as a first-column value" % key)

    def __len__(self):
        return len(self._rows)

    def __contains__(self, key):
        return key in list(self.keys())

    def keys(self):
        """
        Return a list of the first element of each row, which can be
        used for indexing.
        """
        return [ d[0] for (d,f) in self._rows if len(d) > 0 ]

    def has_key(self, key):
        return key in list(self.keys())

    def row(self, key=None, index=None):
        if key is not None:
            if index is not None:
                raise ValueError("Cannot specify *both* key and index")
            for row_data,_ in self._rows:
                if len(row_data) > 0 and row_data[0] == key:
                    return row_data
            raise KeyError("%s not found as a first-column value" % key)

        elif index is not None:
            if 0 <= index < len(self):
                return self._rows[index][0]
            else:
                raise ValueError("Index %d is out of bounds" % index)

        else:
            raise ValueError("Must specify either key or index")


    def col(self, key=None, index=None):
        if key is not None:
            if index is not None:
                raise ValueError("Cannot specify *both* key and index")
            if key in self._columnNames:
                iCol = self._columnNames.index(key)
                return [ d[iCol] for (d,f) in self._rows ] #if len(d)>iCol
            raise KeyError("%s is not a column name." % key)

        elif index is not None:
            if 0 <= index < len(self._columnNames):
                return [ d[index] for (d,f) in self._rows ] #if len(d)>iCol
            else:
                raise ValueError("Index %d is out of bounds" % index)

        else:
            raise ValueError("Must specify either key or index")


    @property
    def num_rows(self):
        return len(self._rows)

    @property
    def num_cols(self):
        return len(self._columnNames)

    @property
    def row_names(self):
        return list(self.keys())

    @property
    def col_names(self):
        return self._columnNames



from __future__ import division, print_function, absolute_import, unicode_literals
#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************
""" GST contraction algorithms """

import numpy as _np
import warnings as _warnings

from .. import objects as _objs
from .. import tools as _tools
from .. import optimize as _opt

def contract(gateset, toWhat, dataset=None, maxiter=1000000, tol=0.01, useDirectCP=True, method="Nelder-Mead", verbosity=0):
    """
    Contract a GateSet to a specified space.

    All contraction operations except 'vSPAM' operate entirely on the gate
    matrices and leave state preparations and measurments alone, while 'vSPAM'
    operations only on SPAM.

    Parameters
    ----------
    gateset : GateSet
        The gateset to contract

    toWhat : string
        Specifies which space is the gateset is contracted to.
        Allowed values are:

        - 'TP'     -- All gates are manifestly trace-preserving maps.
        - 'CP'     -- All gates are manifestly completely-positive maps.
        - 'CPTP'   -- All gates are manifestly completely-positive and trace-preserving maps.
        - 'XP'     -- All gates are manifestly "experimentally-positive" maps.
        - 'XPTP'   -- All gates are manifestly "experimentally-positive" and trace-preserving maps.
        - 'vSPAM'  -- state preparation and measurement operations are valid.
        - 'nothing' -- no contraction is performed.

    dataset : DataSet, optional
        Dataset to use to determine whether a gateset is in the
        "experimentally-positive" (XP) space.  Required only when
        contracting to XP or XPTP.

    maxiter : int, optional
        Maximum number of iterations for iterative contraction routines.

    tol : float, optional
        Tolerance for iterative contraction routines.

    useDirectCP : bool, optional
        Whether to use a faster direct-contraction method for CP
        contraction.  This method essentially transforms to the
        Choi matrix, truncates any negative eigenvalues to zero,
        then transforms back to a gate matrix.

    method : string, optional
        The method used when contracting to XP and non-directly to CP
        (i.e. useDirectCP == False).

    verbosity : int, optional
        How much detail to send to stdout.

    Returns
    -------
    GateSet
        The contracted gateset
    """

    printer = _objs.VerbosityPrinter.build_printer(verbosity)

    if toWhat == 'CPTP':
        if useDirectCP:
            distance,contractedGateset = _contractToCP_direct(gateset, printer, TPalso=True, maxiter=maxiter)
        else:
            distance,contractedGateset = _contractToTP(gateset,verbosity)
            distance,contractedGateset = _contractToCP(contractedGateset, printer, method, maxiter, tol)
    elif toWhat == 'XPTP':
        if dataset is None: raise ValueError("dataset must be given to contract to " + toWhat)
        distance,contractedGateset = _contractToTP(gateset,verbosity)
        distance,contractedGateset = _contractToXP(contractedGateset, dataset,verbosity, method, maxiter, tol)
    elif toWhat == 'CP':
        if useDirectCP:
            distance,contractedGateset = _contractToCP_direct(gateset, printer, TPalso=False, maxiter=maxiter)
        else:
            distance,contractedGateset = _contractToCP(gateset, printer, method, maxiter, tol)
    elif toWhat == 'TP':
        distance,contractedGateset = _contractToTP(gateset,verbosity)
    elif toWhat == 'XP':
        if dataset is None: raise ValueError("dataset must be given to contract to " + toWhat)
        distance,contractedGateset = _contractToXP(gateset,dataset,verbosity,method,maxiter,tol)
    elif toWhat == 'vSPAM':
        contractedGateset = _contractToValidSPAM(gateset, printer)
    elif toWhat == 'nothing':
        contractedGateset = gateset.copy()
    else: raise ValueError("Invalid contract argument: %s" % toWhat)

    return contractedGateset


#modifies gates only (not rhoVecs or EVecs = SPAM)
def _contractToXP(gateset,dataset,verbosity,method='Nelder-Mead',
                 maxiter=100000, tol=1e-10):

    CLIFF = 10000

    printer = _objs.VerbosityPrinter.build_printer(verbosity)

    #printer.log('', 2)
    printer.log("--- Contract to XP ---", 1)
    gs = gateset.copy() #working copy that we keep overwriting with vectorized data

    def objective_func(vectorGS):
        gs.from_vector(vectorGS)
        forbiddenProbPenalty = _tools.forbidden_prob(gs,dataset)
        return (CLIFF + forbiddenProbPenalty if forbiddenProbPenalty > 1e-10 else 0) \
            + gs.frobeniusdist(gateset)

    bToStdout = (printer.verbosity > 2 and printer.filename is None)
    print_obj_func = _opt.create_obj_func_printer(objective_func) #only ever prints to stdout!
    if objective_func(gs.to_vector()) < 1e-8:
        printer.log('Already in XP - no contraction necessary', 1)
        return 0.0, gs

    optSol = _opt.minimize(objective_func,gs.to_vector(),
                          method=method, tol=tol, maxiter=maxiter,
                          callback = print_obj_func if bToStdout else None)

    gs.from_vector(optSol.x)
    #gs.log("Contract to XP", { 'method': method, 'tol': tol, 'maxiter': maxiter } )
    if optSol.fun >= CLIFF: _warnings.warn("Failed to contract gateset to XP")

    printer.log('The closest legal point found was distance: ' + str(optSol.fun), 1)

    return optSol.fun, gs

#modifies gates only (not rhoVecs or EVecs = SPAM)
def _contractToCP(gateset,verbosity,method='Nelder-Mead',
                 maxiter=100000, tol=1e-2):

    CLIFF = 10000
    printer = _objs.VerbosityPrinter.build_printer(verbosity)

    #printer.log('', 2)
    printer.log("--- Contract to CP ---", 1)
    gs = gateset.copy() #working copy that we keep overwriting with vectorized data
    mxBasis = gs.get_basis_name()
    basisDim = gs.get_basis_dimension()

    def objective_func(vectorGS):
        gs.from_vector(vectorGS)
        gs.set_basis(mxBasis,basisDim) #set basis for jamiolkowski iso
        cpPenalty = _tools.sum_of_negative_choi_evals(gs) * 1000
        return (CLIFF + cpPenalty if cpPenalty > 1e-10 else 0) + gs.frobeniusdist(gateset)

    bToStdout = (printer.verbosity > 2 and printer.filename is None)
    print_obj_func = _opt.create_obj_func_printer(objective_func) #only ever prints to stdout!
    if objective_func(gs.to_vector()) < 1e-8:
        printer.log('Already in CP - no contraction necessary', 1)
        return 0.0, gs

    optSol = _opt.minimize(objective_func,gs.to_vector(),
                          method=method, tol=tol, maxiter=maxiter,
                          callback = print_obj_func if bToStdout else None)

    gs.from_vector(optSol.x)
    #gs.log("Contract to CP", { 'method': method, 'tol': tol, 'maxiter': maxiter } )
    if optSol.fun >= CLIFF: _warnings.warn("Failed to contract gateset to CP")

    printer.log('The closest legal point found was distance: ' + str(optSol.fun), 1)

    return optSol.fun, gs


#modifies gates only (not rhoVecs or EVecs = SPAM)
def _contractToCP_direct(gateset,verbosity,TPalso=False,maxiter=100000,tol=1e-8):

    printer = _objs.VerbosityPrinter.build_printer(verbosity)

    gs = gateset.copy() #working copy that we keep overwriting with vectorized data
    mxBasis = gs.get_basis_name()
    #printer.log('', 1)
    printer.log(("--- Contract to %s (direct) ---" % ("CPTP" if TPalso else "CP")), 1)

    for (gateLabel,gate) in gateset.gates.items():
        new_gate = gate.copy()
        if(TPalso):
            for k in range(new_gate.shape[1]): new_gate[0,k] = 1.0 if k == 0 else 0.0

        Jmx = _tools.jamiolkowski_iso(new_gate,gateMxBasis=mxBasis,choiMxBasis="gm")
        evals,evecs = _np.linalg.eig(Jmx)

        if TPalso:
            assert( abs( sum(evals) - 1.0 ) < 1e-8 ) #check that Jmx always has trace == 1
        #if abs( sum(evals) - 1.0 ) >= 1e-8: #DEBUG
        #  print "WARNING: JMx given with evals = %s (sum = %s != 1)" % (evals,sum(evals))
        #  print "WARNING: JMx from: "; _tools.print_mx(new_gate)


        it = 0
        while min(evals) < -tol or abs( sum(evals) - 1.0 ) >= tol:

            #Project eigenvalues to being all positive
            new_evals = evals[:]

            #New projection code
            new_evals = [ max(ev.real,0) for ev in new_evals ]  #don't need .real in theory, but small im parts can snowball in practice
            total_shift = 1.0 - sum(new_evals)  #amount (usually/always < 0) needed to add to eigenvalues to make sum == 1
            sorted_evals_with_inds = sorted( enumerate(new_evals), key=lambda x: x[1] ) # (index,eval) tuples sorted by eval

            shift_left = total_shift
            evals_left = len(sorted_evals_with_inds)
            ideal_shift = shift_left / evals_left

            for (i, sorted_eval) in sorted_evals_with_inds: #loop over new_evals from smallest to largest (note all > 0)
                evals_left -= 1  #number of eigenvalue beyond current eval (in sorted order)
                if sorted_eval+ideal_shift >= 0:
                    new_evals[i] = sorted_eval + ideal_shift
                    shift_left -= ideal_shift
                elif evals_left > 0:
                    new_evals[i] = 0
                    shift_left += sorted_eval
                    ideal_shift = shift_left / evals_left #divide remaining shift evenly among remaining eigenvalues
                else: #last eigenvalue would be < 0 with ideal shift and can't set == 0 b/c all others must be zero too
                    new_evals[i] = 1.0 # so set what was the largest eigenvalue == 1.0

            #if abs( sum(new_evals) - 1.0 ) >= 1e-8:              #DEBUG
            #  print "DEBUG: sum(new_evals) == ",sum(new_evals)   #DEBUG
            #  print "DEBUG: new_evals == ",new_evals             #DEBUG
            #  print "DEBUG: orig evals == ",evals                #DEBUG
            assert( abs( sum(new_evals) - 1.0 ) < 1e-8 )


#  OLD projection code -- can runaway, and take many iters
#        while min(new_evals) < 0:
#          new_evals = [ max(ev.real,0) for ev in new_evals ]  #don't need .real in theory, but small im parts can snowball in practice
#          inds_to_shift = [i for (i,ev) in enumerate(new_evals) if ev > 0]
#          assert(len(inds_to_shift) > 0)
#
#          shift = (1.0 - sum(new_evals))/float(len(inds_to_shift))
#          for i in inds_to_shift: new_evals[i] += shift
#
#          if abs( sum(new_evals) - 1.0 ) >= 1e-8:              #DEBUG
#            print "DEBUG: sum(new_evals) == ",sum(new_evals)   #DEBUG
#            print "DEBUG: new_evals == ",new_evals             #DEBUG
#
#          assert( abs( sum(new_evals) - 1.0 ) < 1e-8 )

            new_Jmx = _np.dot(evecs, _np.dot( _np.diag(new_evals), _np.linalg.inv(evecs) ) )

            #Make trace preserving by zeroing out real parts of off diagonal blocks and imaginary parts
            #  within diagaonal 1x1 and 3x3 block (so really just the 3x3 block's off diag elements)
            #assert(new_Jmx.shape == (4,4)) #NOTE: only works for 1-qubit case so far
            kmax = new_Jmx.shape[0]
            for k in range(1,kmax):
                new_Jmx[0,k] = 1j*new_Jmx[0,k].imag
                new_Jmx[k,0] = 1j*new_Jmx[k,0].imag
            for i in range(1,kmax):
                for j in range(1,kmax):
                    new_Jmx[i,j] = new_Jmx[i,j].real

            evals,evecs = _np.linalg.eig(new_Jmx)

            #DEBUG
            #EVAL_TOL = 1e-10
            #if abs( sum(evals) - 1.0 ) >= 1e-8:
            #  print "DEBUG2: sum(evals) == ",sum(evals)
            #  print "DEBUG2: evals == ",evals
            #if min(evals) < -EVAL_TOL:
            #  print "DEBUG3: evals = ",evals

            assert( min(evals) >= -1e-10 and abs( sum(evals) - 1.0 ) < 1e-8) #Check that trace-trunc above didn't mess up positivity

            new_gate = _tools.jamiolkowski_iso_inv(new_Jmx,gateMxBasis=mxBasis,choiMxBasis="gm")

            #Old way of enforcing TP -- new way should be better since it's not iterative, but keep this around just in case.
            #  new_gate = _tools.jamiolkowski_iso_inv(new_Jmx)
            #
            #  if(TPalso):
            #    for k in range(new_gate.shape[1]):
            #      #if k == 0: assert( abs(new_gate[0,k] - 1.0) < 1e-8 )
            #      #else: assert( abs(new_gate[0,k]) < 1e-8 )
            #      new_gate[0,k] = 1.0 if k == 0 else 0.0
            #
            #  new_Jmx = _tools.jamiolkowski_iso(new_gate)
            #  evals,evecs = _np.linalg.eig(new_Jmx)

            it += 1
            if it > maxiter: break

        gs.gates[gateLabel] = _objs.FullyParameterizedGate( new_gate )

        if it > maxiter:
            printer.warning("Max iterations exceeded in contract_to_cp_direct")
        #else: print "contract_to_cp_direct success in %d iterations" % it  #DEBUG

        printer.log("Direct CP contraction of %s gate gives frobenius diff of %g" % \
              (gateLabel, _tools.frobeniusdist(gs.gates[gateLabel],gate)), 2)

    #gs.log("Choi-Truncate to %s" % ("CPTP" if TPalso else "CP"), { 'maxiter': maxiter } )
    distance = gs.frobeniusdist(gateset)
    printer.log(('The closest legal point found was distance: %s' % str(distance)), 1)

    if TPalso: #TP also constrains prep vectors
        gate_dim = gs.get_dimension()
        for rhoVec in list(gs.preps.values()):
            rhoVec[0,0] = 1.0 / gate_dim**0.25

    return distance, gs


#modifies gates only (not rhoVecs or EVecs = SPAM)
def _contractToTP(gateset,verbosity):
    printer = _objs.VerbosityPrinter.build_printer(verbosity)
    #printer.log('', 2)
    printer.log("--- Contract to TP ---", 1)
    gs = gateset.copy()
    for gate in list(gs.gates.values()):
        gate[0,0] = 1.0
        for k in range(1,gate.shape[1]): gate[0,k] = 0.0

    gate_dim = gs.get_dimension()
    for rhoVec in list(gs.preps.values()):
        rhoVec[0,0] = 1.0 / gate_dim**0.25

    distance = gs.frobeniusdist(gateset)
    printer.log(('Projected TP gateset was at distance: %g' % distance), 1)

    #gs.log("Contract to TP")
    return distance, gs


#modifies rhoVecs and EVecs (SPAM) only (not gates)
def _contractToValidSPAM(gateset, verbosity=0):
    """
    Contract the surface preparation and measurement operations of
    a GateSet to the space of valid quantum operations.

    Parameters
    --------------------
    gateset : GateSet
        The gateset to contract

    verbosity : int
        How much detail to send to stdout.

    Returns
    -------
    GateSet
        The contracted gateset
    """

    printer = _objs.VerbosityPrinter.build_printer(verbosity)

    TOL = 1e-9
    gs = gateset.copy()

    # ** assumption: only the first vector element of pauli vectors has nonzero trace
    dummyVec = _np.zeros( (gateset.get_dimension(),1), 'd'); dummyVec[0,0] = 1.0
    firstElTrace = _np.real( _tools.trace(_tools.ppvec_to_stdmx(dummyVec)))  # == sqrt(2)**nQubits
    diff = 0

    # rhoVec must be positive semidefinite and trace = 1
    for prepLabel,rhoVec in gs.preps.items():
        vec = rhoVec.copy()

        #Ensure trace == 1.0 (maybe later have this be optional)
        firstElTarget = 1.0 / firstElTrace    #TODO: make this function more robust
                                              # to multiple rhovecs -- can only use on ratio,
                                              # so maybe take average of ideal ratios for each rhoVec
                                              # and apply that?  The function works fine now for just one rhovec.
        if abs(firstElTarget - vec[0,0]) > TOL:
            r = firstElTarget / vec[0,0]
            vec *= r  #multiply rhovec by factor
            for ELabel,EVec in gs.effects.items():
                gs.effects[ELabel] = EVec / r

        mx = _tools.ppvec_to_stdmx(vec)

        #Ensure positive semidefinite
        lowEval = min( [ev.real for ev in _np.linalg.eigvals( mx ) ])
        while(lowEval < -TOL):
            idEl = vec[0,0] #only element with trace (even for multiple qubits) -- keep this constant and decrease others
            vec /= 1.00001; vec[0,0] = idEl
            lowEval = min( [ev.real for ev in _np.linalg.eigvals( _tools.ppvec_to_stdmx(vec) ) ])

        diff += _np.linalg.norm( gateset.preps[prepLabel] - vec )
        gs.preps[prepLabel] = vec

    # EVec must have eigenvals between 0 and 1 <==> positive semidefinite and trace <= 1
    for ELabel,EVec in gs.effects.items():
        evals,evecs = _np.linalg.eig( _tools.ppvec_to_stdmx(EVec) )
        if(min(evals) < 0.0 or max(evals) > 1.0):
            if all([ev > 1.0 for ev in evals]):
                evals[ evals.argmin() ] = 0.0 #at least one eigenvalue must be != 1.0
            if all([ev < 0.0 for ev in evals]):
                evals[ evals.argmax() ] = 1.0 #at least one eigenvalue must be != 0.0
            for (k,ev) in enumerate(evals):
                if ev < 0.0: evals[k] = 0.0
                if ev > 1.0: evals[k] = 1.0
            mx = _np.dot(evecs, _np.dot( _np.diag(evals), _np.linalg.inv(evecs) ) )
            vec = _tools.stdmx_to_ppvec(mx)
            diff += _np.linalg.norm( gateset.effects[ELabel] - vec )
            gs.effects[ELabel] = vec

    #gs.log("Contract to valid SPAM")
    #printer.log('', 2)
    printer.log("--- Contract to valid SPAM ---", 1)
    printer.log(("Sum of norm(deltaE) and norm(deltaRho) = %g" % diff), 1)
    for (prepLabel,rhoVec) in gateset.preps.items():
        printer.log("  %s: %s ==> %s " % (prepLabel, str(_np.transpose(rhoVec)),
                                   str(_np.transpose(gs.preps[prepLabel]))), 2)
    for (ELabel,EVec) in gateset.effects.items():
        printer.log("  %s: %s ==> %s " % (ELabel, str(_np.transpose(EVec)),
                                    str(_np.transpose(gs.effects[ELabel]))), 2)

    return gs #return contracted gateset



from __future__ import division, print_function, absolute_import, unicode_literals
#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************
""" Matrix related utility functions """

import numpy as _np
import warnings as _warnings


def trace(M): #memory leak in numpy causes repeated trace calls to eat up all memory --TODO: Cython this
    """
    The trace of a matrix, sum_i M[i,i].

    A memory leak in some version of numpy can cause repeated calls to numpy's
    trace function to eat up all available system memory, and this function
    does not have this problem.

    Parameters
    ----------
    M : numpy array
        the matrix (any object that can be double-indexed)

    Returns
    -------
    element type of M
        The trace of M.
    """
    return sum([ M[i,i] for i in range(M.shape[0]) ])
#    with warnings.catch_warnings():
#        warnings.filterwarnings('error')
#        try:
#            ret =
#        except Warning:
#            print "BAD trace from:\n"
#            for i in range(M.shape[0]):
#                print M[i,i]
#            raise ValueError("STOP")
#    return ret

def is_hermitian(mx, TOL=1e-9):
    """
    Test whether mx is a hermitian matrix.

    Parameters
    ----------
    mx : numpy array
        Matrix to test.

    TOL : float, optional
        Tolerance on absolute magitude of elements.

    Returns
    -------
    bool
        True if mx is hermitian, otherwise False.
    """
    (m,n) = mx.shape
    for i in range(m):
        if abs(mx[i,i].imag) > TOL: return False
        for j in range(i+1,n):
            if abs(mx[i,j] - mx[j,i].conjugate()) > TOL: return False
    return True

def is_pos_def(mx, TOL=1e-9):
    """
    Test whether mx is a positive-definite matrix.

    Parameters
    ----------
    mx : numpy array
        Matrix to test.

    TOL : float, optional
        Tolerance on absolute magitude of elements.

    Returns
    -------
    bool
        True if mx is positive-semidefinite, otherwise False.
    """
    evals = _np.linalg.eigvals( mx )
    return all( [ev > -TOL for ev in evals] )

def is_valid_density_mx(mx, TOL=1e-9):
    """
    Test whether mx is a valid density matrix (hermitian,
      positive-definite, and unit trace).

    Parameters
    ----------
    mx : numpy array
        Matrix to test.

    TOL : float, optional
        Tolerance on absolute magitude of elements.

    Returns
    -------
    bool
        True if mx is a valid density matrix, otherwise False.
    """
    return is_hermitian(mx,TOL) and is_pos_def(mx,TOL) and abs(trace(mx)-1.0) < TOL


def frobeniusnorm(ar):
    """
    Compute the frobenius norm of an array (or matrix),

       sqrt( sum( each_element_of_a^2 ) )

    Parameters
    ----------
    ar : numpy array
        What to compute the frobenius norm of.  Note that ar can be any shape
        or number of dimenions.

    Returns
    -------
    float or complex
        depending on the element type of ar.
    """
    return _np.sqrt(_np.sum(ar**2))

def frobeniusnorm2(ar):
    """
    Compute the squared frobenius norm of an array (or matrix),

       sum( each_element_of_a^2 ) )

    Parameters
    ----------
    ar : numpy array
        What to compute the squared frobenius norm of.  Note that ar can be any
        shape or number of dimenions.

    Returns
    -------
    float or complex
        depending on the element type of ar.
    """
    return _np.sum(ar**2)


def nullspace(m, tol=1e-7):
    """
    Compute the nullspace of a matrix.

    Parameters
    ----------
    m : numpy array
       An matrix of shape (M,N) whose nullspace to compute.

    tol : float (optional)
       Nullspace tolerance, used when comparing singular values with zero.

    Returns
    -------
    An matrix of shape (M,K) whose columns contain nullspace basis vectors.
    """
    _,s,vh = _np.linalg.svd(m)
    rank = (s > tol).sum()
    return vh[rank:].T.copy()


def print_mx(mx, width=9, prec=4):
    """
    Print matrix in pretty format.

    Will print real or complex matrices with a desired precision and
    "cell" width.

    Parameters
    ----------
    mx : numpy array
        the matrix (2-D array) to print.

    width : int, opitonal
        the width (in characters) of each printed element

    prec : int optional
        the precision (in characters) of each printed element

    """
    print(mx_to_string(mx, width, prec))

def mx_to_string(m, width=9, prec=4):
    """
    Generate a "pretty-format" string for a matrix.

    Will generate strings for real or complex matrices with a desired
    precision and "cell" width.

    Parameters
    ----------
    mx : numpy array
        the matrix (2-D array) to convert.

    width : int, opitonal
        the width (in characters) of each converted element

    prec : int optional
        the precision (in characters) of each converted element

    Returns
    -------
    string
        matrix m as a pretty formated string.
    """
    s = ""; tol = 10**(-prec)
    if _np.max(abs(_np.imag(m))) > tol:
        return mx_to_string_complex(m, width, width, prec)

    if len(m.shape) == 1: m = m[None,:] # so it works w/vectors too
    for i in range(m.shape[0]):
        for j in range(m.shape[1]):
            if abs(m[i,j]) < tol: s += '{0: {w}.0f}'.format(0,w=width)
            else: s += '{0: {w}.{p}f}'.format(m[i,j].real,w=width,p=prec)
        s += "\n"
    return s

def mx_to_string_complex(m, real_width=9, im_width=9, prec=4):
    """
    Generate a "pretty-format" string for a complex-valued matrix.

    Parameters
    ----------
    mx : numpy array
        the matrix (2-D array) to convert.

    real_width : int, opitonal
        the width (in characters) of the real part of each element.

    im_width : int, opitonal
        the width (in characters) of the imaginary part of each element.

    prec : int optional
        the precision (in characters) of each element's real and imaginary parts.

    Returns
    -------
    string
        matrix m as a pretty formated string.
    """
    if len(m.shape) == 1: m = m[None,:] # so it works w/vectors too
    s = ""; tol = 10**(-prec)
    for i in range(m.shape[0]):
        for j in range(m.shape[1]):
            if abs(m[i,j].real)<tol: s += "{0: {w}.0f}".format(0,w=real_width)
            else: s += "{0: {w}.{p}f}".format(m[i,j].real,w=real_width,p=prec)
            if abs(m[i,j].imag)<tol: s += "{0: >+{w}.0f}j".format(0,w=im_width)
            else: s += "{0: >+{w}.{p}f}j".format(m[i,j].imag,w=im_width,p=prec)
        s += "\n"
    return s


def real_matrix_log(M, actionIfImaginary="raise",TOL=1e-8):
    """ 
    Construct a *real* logarithm of matrix `M`.

    This is possible when negative eigenvalues of `M` come in pairs, so
    that they can be viewed as complex conjugate pairs.

    Parameters
    ----------
    M : numpy array
        The matrix to take the logarithm of

    actionIfImaginary : {"raise","warn","ignore"}, optional
        What action should be taken if a real-valued logarithm cannot be found.
        "raise" raises a ValueError, "warn" issues a warning, and "ignore"
        ignores the condition and simply returns the complex-valued result.

    TOL : float, optional
        An internal tolerance used when testing for equivalence and zero
        imaginary parts (real-ness).

    Returns
    -------
    numpy array
        An matrix `logM`, of the same shape as `M`, such that `M = exp(logM)`
        and `logM` is real (if possible).
    """
    assert(_np.linalg.norm(_np.imag(M)) < TOL) #M should be real to begin with
    evals,U = _np.linalg.eig(M)

    used_indices = set()
    neg_real_pairs_real_evecs = []
    neg_real_pairs_conj_evecs = []
    unpaired_indices = []
    for i,ev in enumerate(evals):
        if i in used_indices: continue
        used_indices.add(i)
        if abs(_np.imag(ev)) < TOL and _np.real(ev) < 0:
            evec1 = U[:,i]
            if _np.linalg.norm(_np.imag(evec1)) < TOL:
                # evec1 is real, so look for ev2 corresponding to another real evec
                for j,ev2 in enumerate(evals[i+1:],start=i+1):
                    if abs(ev-ev2) < TOL and _np.linalg.norm(_np.imag(U[:,j])) < TOL:
                        used_indices.add(j)
                        neg_real_pairs_real_evecs.append( (i,j) ); break
                else: unpaired_indices.append(i)
            else:
                # evec1 is complex, so look for ev2 corresponding to the conjugate of evec1
                evec1C = evec1.conjugate()
                for j,ev2 in enumerate(evals[i+1:],start=i+1):
                    if abs(ev-ev2) < TOL and _np.linalg.norm(evec1C - U[:,j]) < TOL:
                        used_indices.add(j)
                        neg_real_pairs_conj_evecs.append( (i,j) ); break
                else: unpaired_indices.append(i)

    log_evals = _np.log(evals)
    
    for (i,j) in neg_real_pairs_real_evecs: #need to adjust evecs as well
        log_evals[i] = _np.log(evals[i])
        log_evals[j] = log_evals[i].conjugate()
        U[:,i] = (U[:,i] + 1j*U[:,j])/_np.sqrt(2)
        U[:,j] = U[:,i].conjugate()

    for (i,j) in neg_real_pairs_conj_evecs: # evecs already conjugates of each other
        log_evals[i] = _np.log(evals[i])
        log_evals[j] = log_evals[i].conjugate()

    logM =  _np.dot( U, _np.dot(_np.diag(log_evals), _np.linalg.inv(U) ))
    imMag = _np.linalg.norm(_np.imag(logM))
    #print("DB: log_evals = ",log_evals)
    #print("DEBUG: |imag(logM)| = ",imMag)

    #if there are unpaired negative real eigenvalues, the logarithm might be imaginary
    mayBeImaginary = bool(len(unpaired_indices) > 0)
    
    if mayBeImaginary and imMag > TOL:
        if actionIfImaginary == "raise":
            raise ValueError("Cannot construct a real log: unpaired negative" +
                         " real eigenvalues: %s" % [evals[i] for i in unpaired_indices])
        elif actionIfImaginary == "warn":
            _warnings.warn("Cannot construct a real log: unpaired negative" +
                         " real eigenvalues: %s" % [evals[i] for i in unpaired_indices])
        elif actionIfImaginary == "ignore":
            pass
        else:
            assert(False), "Invalid 'actionIfImaginary' argument: %s" % actionIfImaginary
    else: # logM should be real!
        assert( imMag < TOL ), "real_matrix_log failed to construct a real logarithm (when it should have!)"
        logM = _np.real(logM)
        
    return logM



from __future__ import division, print_function, absolute_import, unicode_literals
#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************
""" Routines for building qutrit gates and gate sets """

import numpy as _np
from scipy import linalg as _linalg

from ..tools import basistools as _bt
from ..      import objects    as _objs


#Define 2 qubit to symmetric (+) antisymmetric space transformation A:
A = _np.matrix([[1,0,0,0],
#               [0,0,0,1],
               [0,1./_np.sqrt(2),1./_np.sqrt(2),0],
               [0,1./_np.sqrt(2),-1./_np.sqrt(2),0],
               [0,0,0,1],])

X = _np.matrix([[0,1],[1,0]])
Y = _np.matrix([[0,-1j],[1j,0]])

#Returns X(theta)^\otimes 2
def X2qubit(theta):
    x = _np.matrix(_linalg.expm(-1j/2. * theta * _np.matrix([[0,1],[1,0]])))
    return _np.kron(x,x)

#Returns Y(theta)^\otimes 2
def Y2qubit(theta):
    y = _np.matrix(_linalg.expm(-1j/2. * theta * _np.matrix([[0,-1j],[1j,0]])))
    return _np.kron(y,y)

#Returns Molmer-Sorensen gate for two qubits
def ms2qubit(theta,phi):
    return _np.matrix(_linalg.expm(-1j/2 * theta *
                                 _np.kron(
                                    _np.cos(phi) * X + _np.sin(phi) * Y,
                                    _np.cos(phi) * X + _np.sin(phi) * Y)
                                 ))

#Projecting above gates into symmetric subspace (qutrit space)
#(state space ordering is |0> = |00>, |1> ~ |01>+|10>,|2>=|11>, so state |i> corresponds to i detector counts


#Removes columns and rows from inputArr
def _remove_from_matrix(inputArr, columns, rows, outputType = _np.matrix):
    inputArr = _np.array(inputArr)
    return outputType([
           [inputArr[row_num][col_num]
           for col_num in range(len(inputArr[row_num]))
           if not col_num in columns]

           for row_num in range(len(inputArr))
           if not row_num in rows])

def to_qutrit_space(inputMat):
    inputMat = _np.matrix(inputMat)
    return _remove_from_matrix(A * inputMat * A**-1,[2],[2])
#    return (A * inputMat * A**-1)[:3,:3]#Comment out above line and uncomment this line if you want the state space
#labelling to be |0>=|00>,|1>=|11>,|2>~|01>+|10>

def MS3(theta,phi):
    return to_qutrit_space(ms2qubit(theta,phi))

def XX3(theta):
    return to_qutrit_space(X2qubit(theta))

def YY3(theta):
    return to_qutrit_space(Y2qubit(theta))

def _random_rot(scale,arrType = _np.array, seed=None):
    rndm = _np.random.RandomState(seed)
    randH = scale * (rndm.randn(3,3) + 1j * rndm.randn(3,3))
    randH = _np.dot(_np.conj(randH.T), randH)
    randU = _linalg.expm(-1j * randH)
    return arrType(randU)

def make_qutrit_gateset(errorScale, Xangle = _np.pi/2, Yangle = _np.pi/2,
                        MSglobal = _np.pi/2, MSlocal = 0,
                        similarity=False,seed=None, basis='gm'):

    arrType = _np.array#Are we casting gates as matrices or arrays?

    rho0 = arrType(([[1,0,0],
                     [0,0,0],
                     [0,0,0]]))

    identity3 = arrType(_np.identity(3))
    identity3final = _bt.change_basis(_np.reshape(identity3,(9,1)), "std", basis)

    E0 = arrType(_np.diag([1,0,0]))
    E1 = arrType(_np.diag([0,1,0]))
    E2 = arrType(_np.diag([0,0,1]))

    #Define gates as unitary ops on Hilbert space
    gateImx = arrType(identity3)
    gateXmx = arrType(XX3(Xangle))
    gateYmx = arrType(YY3(Yangle))
    gateMmx = arrType(MS3(MSglobal,MSlocal))

    #Now introduce unitary noise.

    scale = errorScale
    Xrand = _random_rot(scale,seed = seed)
    Yrand = _random_rot(scale)
    Mrand = _random_rot(scale)
    Irand = _random_rot(scale)


    if similarity:#Change basis for each gate; this preserves rotation angles, and should map identity to identity
        gateXmx = _np.dot(_np.dot(_np.conj(Xrand).T, gateXmx), Xrand)
        gateYmx = _np.dot(_np.dot(_np.conj(Yrand).T, gateYmx), Yrand)
        gateMmx = _np.dot(_np.dot(_np.conj(Mrand).T, gateMmx), Mrand)
        gateImx = _np.dot(_np.dot(_np.conj(Irand).T, gateMmx), Irand)

    else:
        gateXmx = _np.dot(gateXmx, Xrand)
        gateYmx = _np.dot(gateYmx, Yrand)
        gateMmx = _np.dot(gateMmx, Mrand)
        gateImx = _np.dot(gateImx, Mrand)

    #Change gate representation to superoperator in Gell-Mann basis
    gateISO = _np.kron(_np.conj(gateImx),gateImx)
    gateISOfinal = _bt.change_basis(gateISO, "std", basis)
    gateXSO = _np.kron(_np.conj(gateXmx),gateXmx)
    gateXSOfinal = _bt.change_basis(gateXSO, "std", basis)
    gateYSO = _np.kron(_np.conj(gateYmx),gateYmx)
    gateYSOfinal = _bt.change_basis(gateYSO, "std", basis)
    gateMSO = _np.kron(_np.conj(gateMmx),gateMmx)
    gateMSOfinal = _bt.change_basis(gateMSO, "std", basis)

    rho0final = _bt.change_basis(_np.reshape(rho0,(9,1)), "std", basis)
    E0final =  _bt.change_basis(_np.reshape(E0,(9,1)), "std", basis)
    E1final = _bt.change_basis(_np.reshape(E1,(9,1)), "std", basis)
    E2final = _bt.change_basis(_np.reshape(E2,(9,1)), "std", basis)

    qutritGS = _objs.GateSet()
    qutritGS['rho0'] = rho0final
    qutritGS['E0'] = E0final
    qutritGS['E1'] = E1final
    qutritGS['E2'] = E2final
    qutritGS['identity'] = identity3final
    qutritGS.spamdefs['0bright'] = ('rho0','E0')
    qutritGS.spamdefs['1bright'] = ('rho0','E1')
    qutritGS.spamdefs['2bright'] = ('rho0','E2')
    qutritGS['Gi'] = _objs.FullyParameterizedGate(arrType(gateISOfinal))
    qutritGS['Gx'] = _objs.FullyParameterizedGate(arrType(gateXSOfinal))
    qutritGS['Gy'] = _objs.FullyParameterizedGate(arrType(gateYSOfinal))
    qutritGS['Gm'] = _objs.FullyParameterizedGate(arrType(gateMSOfinal))
    qutritGS.set_basis(basis,3)
    
    return qutritGS



from __future__ import division, print_function, absolute_import, unicode_literals
#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************
""" End-to-end functions for performing long-sequence GST """

import os as _os
import warnings as _warnings
import numpy as _np
import sys as _sys
import time as _time
import collections as _collections

from .. import report as _report
from .. import algorithms as _alg
from .. import construction as _construction
from .. import objects as _objs
from .. import io as _io

def do_long_sequence_gst(dataFilenameOrSet, targetGateFilenameOrSet,
                         prepStrsListOrFilename, effectStrsListOrFilename,
                         germsListOrFilename, maxLengths, gaugeOptParams=None,
                         objective="logl", fidPairs=None, lsgstLists=None,
                         advancedOptions=None, comm=None, memLimit=None,
                         verbosity=2):
    """
    Perform end-to-end GST analysis using Ls and germs, with L as a maximum
    length.
    Constructs gate strings by repeating germ strings an integer number of
    times such that the length of the repeated germ is less than or equal to
    the maximum length set in maxLengths.  The LGST estimate of the gates is
    computed, gauge optimized, and then used as the seed for either LSGST or
    MLEGST.
    LSGST is iterated ``len(maxLengths)`` times with successively larger sets
    of gate strings.  On the i-th iteration, the repeated germs sequences
    limited by ``maxLengths[i]`` are included in the growing set of strings
    used by LSGST.  The final iteration will use MLEGST when ``objective ==
    "logl"`` to maximize the true log-likelihood instead of minimizing the
    chi-squared function.
    Once computed, the gate set estimates are optionally gauge optimized to
    the CPTP space and then to the target gate set (using `gaugeOptRatio`
    and `gaugeOptItemWeights`). A :class:`~pygsti.report.Results`
    object is returned, which encapsulates the input and outputs of this GST
    analysis, and can generate final end-user output such as reports and
    presentations.

    Parameters
    ----------
    dataFilenameOrSet : DataSet or string
        The data set object to use for the analysis, specified either directly
        or by the filename of a dataset file (in text format).
    targetGateFilenameOrSet : GateSet or string
        The target gate set, specified either directly or by the filename of a
        gateset file (text format).
    prepStrsListOrFilename : (list of GateStrings) or string
        The state preparation fiducial gate strings, specified either directly
        or by the filename of a gate string list file (text format).
    effectStrsListOrFilename : (list of GateStrings) or string or None
        The measurement fiducial gate strings, specified either directly or by
        the filename of a gate string list file (text format).  If ``None``,
        then use the same strings as specified by prepStrsListOrFilename.
    germsListOrFilename : (list of GateStrings) or string
        The germ gate strings, specified either directly or by the filename of a
        gate string list file (text format).
    maxLengths : list of ints
        List of integers, one per LSGST iteration, which set truncation lengths
        for repeated germ strings.  The list of gate strings for the i-th LSGST
        iteration includes the repeated germs truncated to the L-values *up to*
        and including the i-th one.
        
    gaugeOptParams : dict, optional
        A dictionary of arguments to :func:`gaugeopt_to_target`, specifying
        how the final gauge optimization should be performed.  The keys and
        values of this dictionary may correspond to any of the arguments
        of :func:`gaugeopt_to_target` *except* for the first `gateset` 
        argument, which is specified internally.  The `targetGateset` argument,
        *can* be set, but is specified internally when it isn't.  If `None`,
        then the dictionary `{'itemWeights': {'gates':1.0, 'spam':0.001}}`
        is used.  If `False`, then then *no* gauge optimization is performed.
    objective : {'chi2', 'logl'}, optional
        Specifies which final objective function is used: the chi-squared or
        the log-likelihood.
    fidPairs : list of 2-tuples or dict, optional
        Specifies a subset of all prepStr,effectStr string pairs to be used in
        this analysis.  If `fidPairs` is a list, each element of `fidPairs` is a
        ``(iRhoStr, iEStr)`` 2-tuple of integers, which index a string within
        the state preparation and measurement fiducial strings respectively. If
        `fidPairs` is a dict, then the keys must be germ strings and values are
        lists of 2-tuples as in the previous case.
    lsgstLists : list of gate string lists, optional
        Provides explicit list of gate string lists to be used in analysis; to
        be given if the dataset uses "incomplete" or "reduced" sets of gate
        string.  Default is ``None``.
    advancedOptions : dict, optional
        Specifies advanced options most of which deal with numerical details of
        the objective function or expert-level functionality.  The allowed keys
        and values include:
        - gateLabels = list of strings
        - gsWeights = dict or None
        - starting point = "LGST" (default) or  "target" or GateSet
        - depolarizeStart = float (default == 0)
        - contractStartToCPTP = True / False (default)
        - cptpPenaltyFactor = float (default = 0)
        - tolerance = float
        - maxIterations = int
        - minProbClip = float
        - minProbClipForWeighting = float (default == 1e-4)
        - probClipInterval = tuple (default == (-1e6,1e6)
        - radius = float (default == 1e-4)
        - useFreqWeightedChiSq = True / False (default)
        - nestedGateStringLists = True (default) / False
        - distributeMethod = "gatestrings" or "deriv" (default)
        - profile = int (default == 1)
        - check = True / False (default)
        - truncScheme = "whole germ powers" (default) or "truncated germ powers"
                        or "length as exponent"
    comm : mpi4py.MPI.Comm, optional
        When not ``None``, an MPI communicator for distributing the computation
        across multiple processors.
    memLimit : int or None, optional
        A rough memory limit in bytes which restricts the amount of memory 
        used (per core when run on multi-CPUs).
    verbosity : int, optional
       The 'verbosity' option is an integer specifying the level of 
       detail printed to stdout during the calculation.
       - 0 -- prints nothing
       - 1 -- shows progress bar for entire iterative GST
       - 2 -- show summary details about each individual iteration
       - 3 -- also shows outer iterations of LM algorithm
       - 4 -- also shows inner iterations of LM algorithm
       - 5 -- also shows detailed info from within jacobian
              and objective function calls.

    Returns
    -------
    Results
    """

    #gateLabels : list or tuple
    #    A list or tuple of the gate labels to use when generating the sets of
    #    gate strings used in LSGST iterations.  If ``None``, then the gate
    #    labels of the target gateset will be used.  This option is useful if
    #    you only want to include a *subset* of the available gates in the LSGST
    #    strings (e.g. leaving out the identity gate).
    #
    #weightsDict : dict, optional
    #    A dictionary with ``keys == gate strings`` and ``values ==
    #    multiplicative`` scaling factor for the corresponding gate string. The
    #    default is no weight scaling at all.
    #
    #gaugeOptRatio : float, optional
    #    The ratio spamWeight/gateWeight used for gauge optimizing to the target
    #    gate set.
    #
    #gaugeOptItemWeights : dict, optional
    #   Dictionary of weighting factors for individual gates and spam operators
    #   used during gauge optimization.   Keys can be gate, state preparation,
    #   POVM effect, or spam labels.  Values are floating point numbers.  By
    #   default, gate weights are set to 1.0 and spam weights to gaugeOptRatio.
    #profile : int, optional
    #    Whether or not to perform lightweight timing and memory profiling.
    #    Allowed values are:
    #    
    #    - 0 -- no profiling is performed
    #    - 1 -- profiling enabled, but don't print anything in-line
    #    - 2 -- profiling enabled, and print memory usage at checkpoints
    #truncScheme : str, optional
    #    Truncation scheme used to interpret what the list of maximum lengths
    #    means. If unsure, leave as default. Allowed values are:
    #
    #    - ``'whole germ powers'`` -- germs are repeated an integer number of
    #      times such that the length is less than or equal to the max.
    #    - ``'truncated germ powers'`` -- repeated germ string is truncated
    #      to be exactly equal to the max (partial germ at end is ok).
    #    - ``'length as exponent'`` -- max. length is instead interpreted
    #      as the germ exponent (the number of germ repetitions).





    tRef = _time.time()

    #Note: *don't* specify default dictionary arguments, as this is dangerous
    # because they are mutable objects
    if advancedOptions is None: advancedOptions = {}
    if gaugeOptParams is None: 
        gaugeOptParams = {'itemWeights': {'gates':1.0, 'spam':0.001}}

    profile = advancedOptions.get('profile',1)
    truncScheme = advancedOptions.get('truncScheme',"whole germ powers")

    if profile == 0: profiler = _objs.DummyProfiler()
    elif profile == 1: profiler = _objs.Profiler(comm,False)
    elif profile == 2: profiler = _objs.Profiler(comm,True)
    else: raise ValueError("Invalid value for 'profile' argument (%s)"%profile)

    if 'verbosity' in advancedOptions: #for backward compatibility
        _warnings.warn("'verbosity' as an advanced option is deprecated." +
                       " Please use the 'verbosity' argument directly.")
        verbosity = advancedOptions['verbosity'] 
    if 'memoryLimitInBytes' in advancedOptions: #for backward compatibility
        _warnings.warn("'memoryLimitInBytes' as an advanced option is deprecated." +
                       " Please use the 'memLimit' argument directly.")
        memLimit = advancedOptions['memoryLimitInBytes']

    printer = _objs.VerbosityPrinter.build_printer(verbosity, comm)

    #Get/load target gateset
    if isinstance(targetGateFilenameOrSet, str):
        gs_target = _io.load_gateset(targetGateFilenameOrSet)
    else:
        gs_target = targetGateFilenameOrSet #assume a GateSet object

    #Get/load dataset
    if isinstance(dataFilenameOrSet, str):
        ds = _io.load_dataset(dataFilenameOrSet, True, "aggregate", printer)
        default_dir = _os.path.dirname(dataFilenameOrSet) #default directory for reports, etc
        default_base = _os.path.splitext( _os.path.basename(dataFilenameOrSet) )[0]
    else:
        ds = dataFilenameOrSet #assume a Dataset object
        default_dir = default_base = None

    #Get/load fiducials
    if isinstance(prepStrsListOrFilename, str):
        prepStrs = _io.load_gatestring_list(prepStrsListOrFilename)
    else: prepStrs = prepStrsListOrFilename

    if effectStrsListOrFilename is None:
        effectStrs = prepStrs #use same strings for effectStrs if effectStrsListOrFilename is None
    else:
        if isinstance(effectStrsListOrFilename, str):
            effectStrs = _io.load_gatestring_list(effectStrsListOrFilename)
        else: effectStrs = effectStrsListOrFilename

    #Get/load germs
    if isinstance(germsListOrFilename, str):
        germs = _io.load_gatestring_list(germsListOrFilename)
    else: germs = germsListOrFilename

    #Construct gate sequences
    if lsgstLists is None:
        gateLabels = advancedOptions.get(
            'gateLabels', list(gs_target.gates.keys()))
        nest = advancedOptions.get('nestedGateStringLists',True)
        lsgstLists = _construction.stdlists.make_lsgst_lists(
            gateLabels, prepStrs, effectStrs, germs, maxLengths, fidPairs,
            truncScheme, nest)

    tNxt = _time.time()
    profiler.add_time('do_long_sequence_gst: loading',tRef); tRef=tNxt

    #Starting Point - compute on rank 0 and distribute
    startingPt = advancedOptions.get('starting point',"LGST")
    gate_dim = gs_target.get_dimension()
    if comm is None or comm.Get_rank() == 0:

        #Compute starting point
        if startingPt == "LGST":
            specs = _construction.build_spam_specs(prepStrs=prepStrs, effectStrs=effectStrs,
                                                   prep_labels=gs_target.get_prep_labels(),
                                                   effect_labels=gs_target.get_effect_labels())
            gs_start = _alg.do_lgst(ds, specs, gs_target, svdTruncateTo=gate_dim,
                                    verbosity=printer) # returns a gateset with the *same*
                                                       # parameterizations as gs_target

            #In LGST case, gauge optimimize starting point to the target
            # (historical; sometimes seems to help in practice, since it's gauge
            # optimizing to physical gates (e.g. something in CPTP)
            gs_start = _alg.gaugeopt_to_target(gs_start, gs_target)
              #Note: use *default* gauge-opt params when optimizing

            # Also reset the POVM identity to that of the target.  Essentially,
            # we declare that this basis (gauge) has the same identity as the
            # target (typically the first basis element).
            gs_start.povm_identity = gs_target.povm_identity.copy()

        elif startingPt == "target":
            gs_start = gs_target.copy()
        elif isinstance(startingPt, _objs.GateSet):
            gs_start = startingPt
            startingPt = "User-supplied-GateSet" #for profiler log below
        else:
            raise ValueError("Invalid starting point: %s" % startingPt)
        
        tNxt = _time.time()
        profiler.add_time('do_long_sequence_gst: Starting Point (%s)'
                          % startingPt,tRef); tRef=tNxt                
    
        #Advanced Options can specify further manipulation of starting gate set
        if advancedOptions.get('contractStartToCPTP',False):
            gs_start = _alg.contract(gs_start, "CPTP")
        if advancedOptions.get('depolarizeStart',0) > 0:
            gs_start = gs_start.depolarize(gate_noise=advancedOptions['depolarizeStart'])
    
        if comm is not None: #broadcast starting gate set
            comm.bcast(gs_start, root=0)
    else:
        gs_start = comm.bcast(None, root=0)

    tNxt = _time.time()
    profiler.add_time('do_long_sequence_gst: Prep Initial seed',tRef); tRef=tNxt

    #Run Long-sequence GST on data
    if objective == "chi2":
        gs_lsgst_list = _alg.do_iterative_mc2gst(
            ds, gs_start, lsgstLists,
            tol = advancedOptions.get('tolerance',1e-6),
            cptp_penalty_factor = advancedOptions.get('cptpPenaltyFactor',0),
            maxiter = advancedOptions.get('maxIterations',100000),
            minProbClipForWeighting=advancedOptions.get(
                'minProbClipForWeighting',1e-4),
            probClipInterval = advancedOptions.get(
                'probClipInterval',(-1e6,1e6)),
            returnAll=True, 
            gatestringWeightsDict=advancedOptions.get('gsWeights',None),
            verbosity=printer,
            memLimit=memLimit,
            useFreqWeightedChiSq=advancedOptions.get(
                'useFreqWeightedChiSq',False), profiler=profiler,
            comm=comm, distributeMethod=advancedOptions.get(
                'distributeMethod',"deriv"),
            check_jacobian=advancedOptions.get('check',False),
            check=advancedOptions.get('check',False))

    elif objective == "logl":
        gs_lsgst_list = _alg.do_iterative_mlgst(
          ds, gs_start, lsgstLists,
          tol = advancedOptions.get('tolerance',1e-6),
          cptp_penalty_factor = advancedOptions.get('cptpPenaltyFactor',0),
          maxiter = advancedOptions.get('maxIterations',100000),
          minProbClip = advancedOptions.get('minProbClip',1e-4),
          probClipInterval = advancedOptions.get('probClipInterval',(-1e6,1e6)),
          radius=advancedOptions.get('radius',1e-4),
          returnAll=True, verbosity=printer,
          memLimit=memLimit, profiler=profiler, comm=comm,
          useFreqWeightedChiSq=advancedOptions.get(
                'useFreqWeightedChiSq',False), 
          distributeMethod=advancedOptions.get(
                'distributeMethod',"deriv"),
          check=advancedOptions.get('check',False))
    else:
        raise ValueError("Invalid longSequenceObjective: %s" % objective)

    tNxt = _time.time()
    profiler.add_time('do_long_sequence_gst: total long-seq. opt.',tRef); tRef=tNxt

    #Do final gauge optimization.  gaugeOptParams can be a dict or a list of
    # dicts, each specifying a successive "stage" of gauge optimization.
    go_gs_lsgst_list = gs_lsgst_list
    if gaugeOptParams != False:
        if hasattr(gaugeOptParams,"keys"):
            go_params_list = [gaugeOptParams]
        else: go_params_list = gaugeOptParams

        ordered_go_params_list = []
        for go_params in go_params_list:
            if "targetGateset" not in go_params:
                go_params["targetGateset"] = gs_target

            ordered_go_params_list.append( _collections.OrderedDict( 
                [(k,go_params[k]) for k in sorted(list(go_params.keys()))]))

            go_gs_lsgst_list = [ _alg.gaugeopt_to_target(gs,**go_params)
                                 for gs in go_gs_lsgst_list]

        tNxt = _time.time()
        profiler.add_time('do_long_sequence_gst: gauge optimization',tRef); tRef=tNxt

    truncFn = _construction.stdlists._getTruncFunction(truncScheme)

    ret = _report.Results()
    ret.init_Ls_and_germs(objective, gs_target, ds,
                        gs_start, maxLengths, germs,
                        go_gs_lsgst_list, lsgstLists, prepStrs, effectStrs,
                        truncFn, fidPairs, gs_lsgst_list)
    ret.parameters['minProbClip'] = \
        advancedOptions.get('minProbClip',1e-4)
    ret.parameters['minProbClipForWeighting'] = \
        advancedOptions.get('minProbClipForWeighting',1e-4)
    ret.parameters['probClipInterval'] = \
        advancedOptions.get('probClipInterval',(-1e6,1e6))
    ret.parameters['radius'] = advancedOptions.get('radius',1e-4)
    ret.parameters['weights'] = advancedOptions.get('gsWeights',None)
    ret.parameters['defaultDirectory'] = default_dir
    ret.parameters['defaultBasename'] = default_base
    ret.parameters['memLimit'] = memLimit
    ret.parameters['gaugeOptParams'] = ordered_go_params_list
    ret.parameters['cptpPenaltyFactor'] = advancedOptions.get('cptpPenaltyFactor',0)
    ret.parameters['distributeMethod'] = advancedOptions.get('distributeMethod','deriv')
    ret.parameters['starting point'] = startingPt
    ret.parameters['depolarizeStart'] = advancedOptions.get('depolarizeStart',0)
    ret.parameters['contractStartToCPTP'] = advancedOptions.get('contractStartToCPTP',False)
    ret.parameters['tolerance'] = advancedOptions.get('tolerance',1e-6)
    ret.parameters['maxIterations'] = advancedOptions.get('maxIterations',100000)
    ret.parameters['useFreqWeightedChiSq'] = advancedOptions.get('useFreqWeightedChiSq',False)
    ret.parameters['nestedGateStringLists'] = advancedOptions.get('nestedGateStringLists',True)
    ret.parameters['profile'] = advancedOptions.get('profile',1)
    ret.parameters['check'] = advancedOptions.get('check',False)
    ret.parameters['truncScheme'] = advancedOptions.get('truncScheme', "whole germ powers")

    profiler.add_time('do_long_sequence_gst: results initialization',tRef)
    ret.parameters['profiler'] = profiler

    assert( len(maxLengths) == len(lsgstLists) == len(go_gs_lsgst_list) )
    return ret
